# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer lithium strategy definitions."""

from abc import ABC
from logging import getLogger

from lithium.strategies import CheckOnly
from lithium.strategies import CollapseEmptyBraces as LithCollapseEmptyBraces
from lithium.strategies import Minimize
from lithium.strategies import Strategy as LithStrategy
from lithium.testcases import Testcase as LithTestcase
from lithium.testcases import TestcaseAttrs, TestcaseChar, TestcaseJsStr, TestcaseLine

from ...common.storage import TestCase
from . import Strategy, _contains_dd

LOG = getLogger(__name__)


class _LithiumStrategy(Strategy, ABC):
    """Use a Lithium `Strategy`/`Testcase` pair to reduce the given Grizzly `TestCase`
    set.

    Implementors must define these class attributes:

    Class attributes:
        name (str): The strategy name.
        strategy_cls (lithium.strategies.Strategy): Lithium strategy type.
        testcase_cls (lithium.testcases.Testcase): Lithium testcase type.
    """

    strategy_cls = None
    testcase_cls = None

    def __init__(self, testcases):
        """Initialize strategy instance.

        Arguments:
            testcases (list(grizzly.common.storage.TestCase)):
                List of testcases to reduce. The object does not take ownership of the
                testcases.
        """
        super().__init__(testcases)
        self._current_reducer = None
        self._files_to_reduce = []
        self.rescan_files_to_reduce()
        self._current_feedback = None
        self._current_served = None

    def rescan_files_to_reduce(self):
        """Repopulate the private `files_to_reduce` attribute by scanning the testcase
        root.

        Returns:
            None
        """
        self._files_to_reduce.clear()
        for path in self._testcase_root.glob("**/*"):
            if path.is_file() and path.name not in {"test_info.json", "prefs.js"}:
                if _contains_dd(path):
                    self._files_to_reduce.append(path)

    @classmethod
    def sanity_check_cls_attrs(cls):
        """Sanity check the strategy class implementation.

        Raises:
            AssertionError: Required class attributes are missing or wrong type.

        Returns:
            None
        """
        super().sanity_check_cls_attrs()
        assert issubclass(cls.strategy_cls, LithStrategy)
        assert issubclass(cls.testcase_cls, LithTestcase)

    def update(self, success, served=None):
        """Inform the strategy whether or not the last reduction yielded was good.

        Arguments:
            success (bool): Whether or not the last reduction was acceptable.
            served (list(list(str))): The list of served files for each testcase in the
                                      last reduction.

        Returns:
            None
        """
        if self._current_reducer is not None:
            self._current_reducer.feedback(success)
        self._current_feedback = success
        self._current_served = served

    def __iter__(self):
        """Iterate over potential reductions of testcases according to this strategy.

        The caller should evaluate each testcase set yielded, and call `update` with the
        result. The caller owns the testcases yielded, and should call `cleanup` for
        each.

        Yields:
            list(grizzly.common.storage.TestCase): list of testcases with reduction
                                                   applied
        """
        LOG.info("Reducing %d files", len(self._files_to_reduce))
        file_no = 0
        reduce_queue = self._files_to_reduce.copy()
        reduce_queue.sort()  # not necessary, but helps make tests more predictable
        # indicates that self._testcase_root contains changes that haven't been yielded
        # (if iteration ends, changes would be lost)
        testcase_root_dirty = False
        while reduce_queue:
            LOG.debug("Reduce queue: %r", reduce_queue)
            file = reduce_queue.pop(0)
            file_no += 1
            LOG.info(
                "[%s] Reducing %s (file %d/%d)",
                self.name,
                file.relative_to(self._testcase_root),
                file_no,
                len(self._files_to_reduce),
            )
            lithium_testcase = self.testcase_cls()  # pylint: disable=not-callable
            lithium_testcase.load(file)
            strategy = self.strategy_cls()  # pylint: disable=not-callable
            self._current_reducer = strategy.reduce(lithium_testcase)

            # populate the lithium strategy "tried" cache
            # use all cache values where all hashes other than the current file match
            # the current testcase_root state.
            current_tc_hash_map = dict(self._calculate_testcase_hash())
            del current_tc_hash_map[str(file.relative_to(self._testcase_root))]
            this_tc_tried = set()
            for tried in self._tried:
                tried = dict(tried)
                tc_tried = tried.pop(str(file.relative_to(self._testcase_root)))
                if tried == current_tc_hash_map:
                    this_tc_tried.add(tc_tried)
            self._current_reducer.update_tried(this_tc_tried)

            for reduction in self._current_reducer:
                reduction.dump()
                testcases = TestCase.load(str(self._testcase_root), True)
                LOG.info("[%s] %s", self.name, self._current_reducer.description)
                yield testcases
                if self._current_feedback:
                    testcase_root_dirty = False
                else:
                    self._tried.add(self._calculate_testcase_hash())
                if self._current_feedback and self._current_served is not None:
                    testcases = TestCase.load(str(self._testcase_root), True)
                    try:
                        self.purge_unserved(testcases, self._current_served)
                    finally:
                        for testcase in testcases:
                            testcase.cleanup()
                    num_files_before = len(self._files_to_reduce)
                    LOG.debug("files being reduced before: %r", self._files_to_reduce)
                    self.rescan_files_to_reduce()
                    LOG.debug("files being reduced after: %r", self._files_to_reduce)
                    files_to_reduce = set(self._files_to_reduce)
                    reduce_queue = list(sorted(set(reduce_queue) & files_to_reduce))
                    testcase_root_dirty = len(self._files_to_reduce) != num_files_before
                    if file not in files_to_reduce:
                        # current reduction was for a purged file
                        break
            else:
                # write out the best found testcase
                self._current_reducer.testcase.dump()
            self._current_reducer = None
        if testcase_root_dirty:
            # purging unserved files enabled us to exit early from the loop.
            # need to yield once more to set this trimmed version to the current best
            # in ReduceManager
            testcases = TestCase.load(str(self._testcase_root), True)
            LOG.info("[%s] final iteration triggered by purge_optional", self.name)
            yield testcases
            assert self._current_feedback, "Purging unserved files broke the testcase."


class Check(_LithiumStrategy):
    """Check whether the testcase is reproducible.

    This strategy does no reduction, and only yields once. It is intended to provide a
    pass/fail result in a reduction pipeline.
    """

    name = "check"
    strategy_cls = CheckOnly
    testcase_cls = TestcaseLine

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        # trim files_to_reduce, for check we don't need to run on every file
        # just once per Grizzly TestCase set is enough.
        self._files_to_reduce = self._files_to_reduce[:1]

    def __iter__(self):
        yield from super().__iter__()


class CollapseEmptyBraces(_LithiumStrategy):
    """Minimize lines, but collapse empty curly braces between each iteration.

    During reduction, the contents of a block may be reduced away entirely, but removing
    the starting brace or ending brace alone will break the syntax of the test file.
    This strategy tries to collapse empty braces onto the same line between each
    iteration, so that empty blocks can be removed if otherwise possible.
    """

    name = "collapsebraces"
    strategy_cls = LithCollapseEmptyBraces
    testcase_cls = TestcaseLine


class MinimizeChars(_LithiumStrategy):
    """Minimize all bytes in the testcase."""

    name = "chars"
    strategy_cls = Minimize
    testcase_cls = TestcaseChar


class MinimizeJSChars(_LithiumStrategy):
    """Minimize all bytes contained in javascript strings.

    This works the same as MinimizeChars, but only operates if it can identify what
    looks like a quoted string. It also treats escaped characters as a single token
    for reduction.
    """

    name = "jschars"
    strategy_cls = Minimize
    testcase_cls = TestcaseJsStr


class MinimizeLines(_LithiumStrategy):
    """Minimize all lines in the testcase."""

    name = "lines"
    strategy_cls = Minimize
    testcase_cls = TestcaseLine


class MinimizeAttrs(_LithiumStrategy):
    """Remove HTML/XML attributes and their values."""

    name = "attrs"
    strategy_cls = Minimize
    testcase_cls = TestcaseAttrs



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from itertools import count
from os.path import join as pathjoin

from pytest import mark, raises

from sapphire import (
    SERVED_ALL,
    SERVED_NONE,
    SERVED_REQUEST,
    SERVED_TIMEOUT,
    Sapphire,
    ServerMap,
)

from ..target import Target, TargetLaunchError, TargetLaunchTimeout
from .reporter import Report
from .runner import Runner, RunResult, _IdleChecker
from .storage import TestCase


def test_runner_01(mocker, tmp_path):
    """test Runner()"""
    mocker.patch("grizzly.common.runner.time", autospec=True, side_effect=count())
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    runner = Runner(server, target, relaunch=10)
    assert runner._idle is None
    assert runner._relaunch == 10
    assert runner._tests_run == 0
    serv_files = ["a.bin", "/another/file.bin"]
    testcase = mocker.Mock(spec=TestCase, landing_page=serv_files[0], optional=[])
    # all files served
    serv_map = ServerMap()
    server.serve_path.return_value = (SERVED_ALL, serv_files)
    result = runner.run([], serv_map, testcase)
    assert result.attempted
    assert runner._tests_run == 1
    assert result.duration == 1
    assert result.initial
    assert result.status is None
    assert result.served == serv_files
    assert not result.timeout
    assert not serv_map.dynamic
    assert target.close.call_count == 0
    assert target.dump_coverage.call_count == 0
    assert target.handle_hang.call_count == 0
    assert testcase.dump.call_count == 1
    # dump coverage
    serv_map = ServerMap()
    server.serve_path.return_value = (SERVED_ALL, serv_files)
    result = runner.run([], serv_map, testcase, coverage=True)
    assert result.attempted
    assert runner._tests_run == 2
    assert not result.initial
    assert result.status is None
    assert result.served == serv_files
    assert not result.timeout
    assert not serv_map.dynamic
    assert target.close.call_count == 0
    assert target.dump_coverage.call_count == 1
    assert target.handle_hang.call_count == 0
    # existing test path
    testcase.reset_mock()
    tc_path = tmp_path / "tc"
    tc_path.mkdir()
    serv_map = ServerMap()
    server.serve_path.return_value = (SERVED_ALL, serv_files)
    result = runner.run([], serv_map, testcase, test_path=str(tc_path))
    assert result.attempted
    assert runner._tests_run == 3
    assert not result.initial
    assert result.status is None
    assert not serv_map.dynamic
    assert target.close.call_count == 0
    assert testcase.dump.call_count == 0
    tc_path.is_dir()


def test_runner_02(mocker):
    """test Runner.run() relaunch"""
    mocker.patch("grizzly.common.runner.time", autospec=True, return_value=1)
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    serv_files = ["a.bin"]
    server.serve_path.return_value = (SERVED_ALL, serv_files)
    testcase = mocker.Mock(spec=TestCase, landing_page=serv_files[0], optional=[])
    # single run/iteration relaunch (not idle exit)
    target.is_idle.return_value = False
    runner = Runner(server, target, relaunch=1)
    assert runner._relaunch == 1
    smap = ServerMap()
    result = runner.run([], smap, testcase)
    assert result.attempted
    assert result.initial
    assert target.close.call_count == 1
    assert target.is_idle.call_count > 0
    assert target.monitor.is_healthy.call_count > 0
    assert result.status is None
    assert result.served == serv_files
    assert not smap.dynamic
    assert smap.redirect.get("grz_next_test").target == "grz_empty"
    assert not result.timeout
    target.reset_mock()
    testcase.reset_mock()
    # single run/iteration relaunch (idle exit)
    target.is_idle.return_value = True
    runner = Runner(server, target, relaunch=1)
    assert runner._relaunch == 1
    result = runner.run([], ServerMap(), testcase)
    assert result.attempted
    assert target.close.call_count == 1
    assert target.monitor.is_healthy.call_count > 0
    target.reset_mock()
    testcase.reset_mock()
    # multiple runs/iterations relaunch (is_healty exit)
    runner = Runner(server, target, relaunch=3)
    target.monitor.is_healthy.return_value = False
    for _ in range(2):
        smap = ServerMap()
        result = runner.run([], smap, testcase)
        assert result.attempted
        assert target.close.call_count == 0
        assert target.monitor.is_healthy.call_count == 0
        assert result.status is None
        assert result.served == serv_files
        assert not result.timeout
        assert not smap.dynamic
        assert "grz_next_test" not in smap.redirect
    smap = ServerMap()
    result = runner.run([], smap, testcase)
    assert runner._tests_run == 3
    assert result.attempted
    assert target.close.call_count == 1
    assert target.is_idle.call_count == 0
    assert target.monitor.is_healthy.call_count == 1
    assert result.status is None
    assert result.served == serv_files
    assert not smap.dynamic
    assert smap.redirect.get("grz_next_test").target == "grz_empty"


@mark.parametrize(
    "srv_result, served",
    [
        # no files served
        (SERVED_NONE, []),
        # landing page not served
        (SERVED_REQUEST, ["harness"]),
    ],
)
def test_runner_03(mocker, srv_result, served):
    """test Runner() errors"""
    server = mocker.Mock(spec=Sapphire)
    server.serve_path.return_value = (srv_result, served)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    testcase = mocker.Mock(spec=TestCase, landing_page="x", optional=[])
    runner = Runner(server, target)
    result = runner.run([], ServerMap(), testcase)
    assert result.status is None
    assert not result.attempted
    assert result.initial
    assert set(result.served) == set(served)
    assert not result.timeout
    assert target.close.call_count == 1


@mark.parametrize(
    "ignore, status, idle, detect_failure",
    [
        # detect a hang
        (["memory"], RunResult.FAILED, False, 1),
        # ignore a hang
        (["timeout"], RunResult.IGNORED, False, 0),
        # ignore idle hang
        ([], RunResult.IGNORED, True, 0),
    ],
)
def test_runner_04(mocker, ignore, status, idle, detect_failure):
    """test reporting timeout"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    testcase = mocker.Mock(spec=TestCase, landing_page="a.bin", optional=[])
    serv_files = ["a.bin", "/another/file.bin"]
    server.serve_path.return_value = (SERVED_TIMEOUT, serv_files)
    target.detect_failure.return_value = target.RESULT_FAILURE
    target.handle_hang.return_value = idle
    target.monitor.is_healthy.return_value = False
    runner = Runner(server, target, relaunch=1)
    serv_map = ServerMap()
    result = runner.run(ignore, serv_map, testcase)
    assert result.status == status
    assert result.served == serv_files
    assert result.timeout
    assert "grz_empty" not in serv_map.dynamic
    assert target.detect_failure.call_count == detect_failure
    assert target.handle_hang.call_count == 1


@mark.parametrize(
    "served, attempted, target_result, status",
    [
        # FAILURE
        (["a.bin"], True, Target.RESULT_FAILURE, RunResult.FAILED),
        # IGNORED
        (["a.bin"], True, Target.RESULT_IGNORED, RunResult.IGNORED),
        # failure before serving landing page
        (["harness"], False, Target.RESULT_FAILURE, RunResult.FAILED),
    ],
)
def test_runner_05(mocker, served, attempted, target_result, status):
    """test reporting failures"""
    server = mocker.Mock(spec=Sapphire)
    server.serve_path.return_value = (SERVED_REQUEST, served)
    target = mocker.Mock(spec=Target, launch_timeout=10)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_IGNORED = Target.RESULT_IGNORED
    target.detect_failure.return_value = target_result
    target.monitor.is_healthy.return_value = False
    testcase = mocker.Mock(spec=TestCase, landing_page="a.bin", optional=[])
    runner = Runner(server, target)
    runner.launch("http://a/")
    result = runner.run([], ServerMap(), testcase)
    assert result.attempted == attempted
    assert result.status == status
    assert not result.timeout
    assert target.handle_hang.call_count == 0
    assert target.close.call_count == 1


def test_runner_06(mocker):
    """test Runner() with idle checking"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    serv_files = ["/fake/file", "/another/file.bin"]
    server.serve_path.return_value = (SERVED_ALL, serv_files)
    runner = Runner(server, target, idle_threshold=0.01, idle_delay=0.01, relaunch=10)
    assert runner._idle is not None
    result = runner.run(
        [],
        ServerMap(),
        mocker.Mock(spec=TestCase, landing_page=serv_files[0], optional=[]),
    )
    assert result.status is None
    assert result.attempted
    assert target.close.call_count == 0


def test_runner_07(mocker):
    """test Runner._keep_waiting()"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)

    target.monitor.is_healthy.return_value = True
    runner = Runner(server, target)
    assert runner._keep_waiting()

    target.monitor.is_healthy.return_value = False
    assert not runner._keep_waiting()

    runner._idle = mocker.Mock(spec=_IdleChecker)
    runner._idle.is_idle.return_value = False
    target.monitor.is_healthy.return_value = True
    assert runner._keep_waiting()

    runner._idle.is_idle.return_value = True
    target.monitor.is_healthy.return_value = True
    assert not runner._keep_waiting()

    runner._idle.is_idle.return_value = False
    target.monitor.is_healthy.return_value = False
    assert not runner._keep_waiting()


def test_runner_08():
    """test Runner.location()"""
    result = Runner.location("a.html", 34567)
    assert result == "http://127.0.0.1:34567/a.html"
    result = Runner.location("/a.html", 34567)
    assert result == "http://127.0.0.1:34567/a.html"
    result = Runner.location("a.html", 34567, close_after=10)
    assert result == "http://127.0.0.1:34567/a.html?close_after=10"
    result = Runner.location("a.html", 9999, time_limit=60)
    assert result == "http://127.0.0.1:9999/a.html?time_limit=60000"
    result = Runner.location("a.html", 9999, close_after=10, time_limit=60)
    assert result == "http://127.0.0.1:9999/a.html?close_after=10&time_limit=60000"


def test_runner_09(mocker):
    """test Runner.launch()"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, launch_timeout=30)
    runner = Runner(server, target)
    # successful launch
    runner._tests_run = 1
    runner.launch("http://a/")
    assert runner._tests_run == 0
    assert target.launch.call_count == 1
    target.reset_mock()
    # target launch error
    target.launch.side_effect = TargetLaunchError("test", mocker.Mock(spec=Report))
    with raises(TargetLaunchError, match="test"):
        runner.launch("http://a/")
    assert target.launch.call_count == 3
    target.reset_mock()
    # target launch timeout
    target.launch.side_effect = TargetLaunchTimeout
    with raises(TargetLaunchTimeout):
        runner.launch("http://a/", max_retries=3)
    assert target.launch.call_count == 3


def test_runner_10(mocker, tmp_path):
    """test Runner.run() adding includes to testcase"""
    server = mocker.Mock(spec=Sapphire)
    target = mocker.Mock(spec=Target)
    target.detect_failure.return_value = target.RESULT_NONE
    runner = Runner(server, target, relaunch=10)
    # create test files
    inc_path1 = tmp_path / "include"
    inc_path1.mkdir()
    inc1 = inc_path1 / "inc_file.bin"
    inc1.write_bytes(b"a")
    (inc_path1 / "nested").mkdir()
    inc2 = inc_path1 / "nested" / "nested_inc.bin"
    inc2.write_bytes(b"a")
    inc_path2 = tmp_path / "include2"
    inc_path2.mkdir()
    inc3 = inc_path2 / "inc_file3.txt"
    inc3.write_bytes(b"a")
    # build server map
    smap = ServerMap()
    smap.set_include("/", str(inc_path1))
    smap.set_include("/test", str(inc_path2))
    serv_files = ["a.b", str(inc1), str(inc2), str(inc3)]
    server.serve_path.return_value = (SERVED_ALL, serv_files)
    with TestCase("a.b", "x", "x") as tcase:
        result = runner.run([], smap, tcase)
        assert result.attempted
        assert result.status is None
        assert "inc_file.bin" in tcase._existing_paths
        assert pathjoin("nested", "nested_inc.bin") in tcase._existing_paths
        assert pathjoin("test", "inc_file3.txt") in tcase._existing_paths


def test_idle_check_01(mocker):
    """test simple _IdleChecker"""
    fake_time = mocker.patch("grizzly.common.runner.time", autospec=True)
    ichk = _IdleChecker(mocker.Mock(), 95, 10, poll_delay=1)
    assert ichk._threshold == 95
    assert ichk._init_delay == 10
    assert ichk._poll_delay == 1
    assert ichk._next_poll is None
    fake_time.return_value = 0
    ichk.schedule_poll(initial=True)
    assert ichk._next_poll == 10
    ichk.schedule_poll()
    assert ichk._next_poll == 1


def test_idle_check_02(mocker):
    """test _IdleChecker.is_idle()"""
    fake_time = mocker.patch("grizzly.common.runner.time", autospec=True)
    callbk = mocker.Mock()
    callbk.return_value = False
    ichk = _IdleChecker(callbk, 99, 10, poll_delay=1)
    fake_time.return_value = 0
    ichk.schedule_poll()
    # early check
    assert not ichk.is_idle()
    assert callbk.call_count == 0
    # not idle
    fake_time.return_value = 10
    assert not ichk.is_idle()
    assert ichk._next_poll == 11
    assert callbk.call_count == 1
    # idle
    callbk.return_value = True
    fake_time.return_value = ichk._next_poll
    assert ichk.is_idle()
    assert callbk.call_count == 2



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from argparse import ArgumentParser

from loki import Loki

__author__ = "Tyson Smith"


def parse_args(argv=None):
    parser = ArgumentParser(description="Loki fuzzing library")
    parser.add_argument("input", help="Output will be generated based on this file")
    parser.add_argument(
        "-a",
        "--aggression",
        default=0.001,
        type=float,
        help="Maximum fuzz rate. 1.0 == 100%% (default: %(default)s)",
    )
    parser.add_argument(
        "-b",
        "--byte-order",
        default=None,
        help="Byte order to use when mutating multiple bytes at once. "
        "Use '>' for big-endian or '<' for little-endian (default: random)",
    )
    parser.add_argument(
        "-c",
        "--count",
        default=1,
        type=int,
        help="Number test cases to generate, minimum 1 (default: %(default)s)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        default=False,
        action="store_true",
        help="Display limited output (default: %(default)s)",
    )
    parser.add_argument(
        "-o",
        "--output",
        default=None,
        help="Output directory for fuzzed test cases (default: '.')",
    )
    args = parser.parse_args(argv)

    if args.byte_order and args.byte_order not in Loki.BYTE_ORDERS:
        parser.error("Invalid byte order %r" % (args.byte_order,))

    return args



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from logging import getLogger
from os.path import dirname
from os.path import join as pathjoin
from shutil import rmtree
from tempfile import mkdtemp

from FTB.Signatures.CrashInfo import CrashSignature

from sapphire import Sapphire, ServerMap

from ..common.plugins import load as load_plugin
from ..common.reporter import FilesystemReporter, FuzzManagerReporter, Report
from ..common.runner import Runner, RunResult
from ..common.status import Status
from ..common.storage import TestCase, TestCaseLoadFailure, TestFile
from ..common.utils import TIMEOUT_DELAY, ConfigError, configure_logging, grz_tmp
from ..session import Session
from ..target import Target, TargetLaunchError, TargetLaunchTimeout

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class ReplayResult:
    __slots__ = ("count", "durations", "expected", "report", "served")

    def __init__(self, report, served, durations, expected):
        self.count = 1
        self.durations = durations
        self.expected = expected
        self.report = report
        self.served = served


class ReplayManager:
    HARNESS_FILE = pathjoin(dirname(__file__), "..", "common", "harness.html")
    DEFAULT_TIME_LIMIT = 30

    __slots__ = (
        "ignore",
        "server",
        "status",
        "target",
        "_any_crash",
        "_harness",
        "_signature",
        "_relaunch",
        "_unpacked",
    )

    def __init__(
        self,
        ignore,
        server,
        target,
        any_crash=False,
        relaunch=1,
        signature=None,
        use_harness=True,
    ):
        self.ignore = ignore
        self.server = server
        self.status = None
        self.target = target
        self._any_crash = any_crash
        self._harness = None
        self._relaunch = relaunch
        self._signature = signature
        if use_harness:
            with open(self.HARNESS_FILE, "rb") as in_fp:
                self._harness = in_fp.read()
        else:
            # target must relaunch every iteration when not using harness
            assert relaunch == 1

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    @property
    def signature(self):
        return self._signature

    @staticmethod
    def check_match(signature, report, expect_hang):
        if signature is None:
            return False
        if expect_hang and not report.is_hang:
            # avoid catching other crashes with forgiving hang signatures
            return False
        return signature.matches(report.crash_info)

    def cleanup(self):
        """Remove temporary files from disk.

        Args:
            None

        Returns:
            None
        """
        if self.status is not None:
            self.status.cleanup()

    @staticmethod
    def expect_hang(ignore, signature, tests):
        """Check if any test is expected to trigger a hang. If a hang is expected
        a sanity check is performed. A ConfigError is raised if a configuration
        issue is detected.

        Args:
            ignore (list(str)): Failure types to ignore.
            signature (CrashSignature): Signature to use for bucketing.
            tests list(TestCase): Testcases to check.

        Returns:
            bool: True if a hang is expected otherwise False.
        """
        is_hang = any(x.hang for x in tests)
        if is_hang:
            if signature is None:
                raise ConfigError(
                    "Hangs require a signature to replay", Session.EXIT_ERROR
                )
            if "timeout" in ignore:
                raise ConfigError(
                    "Cannot ignore 'timeout' when detecting hangs", Session.EXIT_ERROR
                )
        return is_hang

    @classmethod
    def load_testcases(cls, path, load_prefs, subset=None):
        """Load TestCases.

        Args:
            path (str): Path to load.
            load_prefs (bool): Load prefs.js file if available.
            subset (list(int)): Indices of tests to load when loading multiple
                                tests.
        Returns:
            list(TestCase): Loaded TestCases.
        """
        LOG.debug("loading the TestCases")
        testcases = TestCase.load(path, load_prefs)
        if not testcases:
            raise TestCaseLoadFailure("Failed to load TestCases")
        if subset:
            count = len(testcases)
            # deduplicate and limit requested indices to valid range
            reqs = {max(count + x, 0) if x < 0 else min(x, count - 1) for x in subset}
            LOG.debug("using TestCase(s) with index %r", reqs)
            selected = list()
            for idx in sorted(reqs, reverse=True):
                selected.append(testcases.pop(idx))
            selected.reverse()
            for test in testcases:
                test.cleanup()
            testcases = selected
        return testcases

    @staticmethod
    def report_to_filesystem(path, results, tests=None):
        """Use FilesystemReporter to write reports and testcase to disk in a
        known location.

        Args:
            path (str): Location to write data.
            results (iterable): ReplayResult to output.
            tests (iterable): Testcases to output.

        Returns:
            None
        """
        others = list(x.report for x in results if not x.expected)
        if others:
            reporter = FilesystemReporter(
                pathjoin(path, "other_reports"), major_bucket=False
            )
            for report in others:
                reporter.submit(tests or [], report=report)
        expected = list(x for x in results if x.expected)
        if expected:
            if tests and len(expected) == 1:
                # only purge optional is reporting a single testcase
                assert len(tests) >= len(expected[0].served)
                for test, served in zip(tests, expected[0].served):
                    LOG.debug("calling test.purge_optional() with %r", served)
                    test.purge_optional(served)
            reporter = FilesystemReporter(pathjoin(path, "reports"), major_bucket=False)
            for result in expected:
                reporter.submit(tests or [], report=result.report)

    def run(
        self,
        testcases,
        time_limit,
        repeat=1,
        min_results=1,
        exit_early=True,
        expect_hang=False,
        idle_delay=0,
        idle_threshold=0,
    ):
        """Run testcase replay.

        Args:
            testcases (list): One or more TestCases to run.
            time_limit (int): Maximum time in seconds a test should take to complete.
            repeat (int): Maximum number of times to run the TestCase.
            min_results (int): Minimum number of results needed before run can
                               be considered successful.
            expect_hang (bool): Running testcases is expected to result in a hang.
            exit_early (bool): If True the minimum required number of iterations
                               are performed to either meet `min_results` or
                               determine that it is not possible to do so.
                               If False `repeat` number of iterations are
                               performed.
            idle_delay (int): Number of seconds to wait before polling for idle.
            idle_threshold (int): CPU usage threshold to mark the process as idle.

        Returns:
            list: ReplayResults that were found running testcases.
        """
        assert idle_delay >= 0
        assert idle_threshold >= 0
        assert min_results > 0
        assert repeat > 0
        assert repeat >= min_results
        assert testcases
        assert time_limit > 0
        assert self._harness is not None or len(testcases) == 1
        assert not expect_hang or self._signature is not None

        if self.status is not None:
            LOG.debug("clearing previous status data")
            self.status.cleanup()
        self.status = Status.start()

        server_map = ServerMap()
        if self._harness is not None:
            server_map.set_dynamic_response(
                "grz_harness", lambda: self._harness, mime_type="text/html"
            )

        # track unprocessed results
        reports = dict()
        # track unpacked testcases
        unpacked = list()
        try:
            sig_hash = Report.calc_hash(self._signature) if self._signature else None
            test_count = len(testcases)
            LOG.debug("unpacking testcases (%d)...", test_count)
            for test in testcases:
                dst_path = mkdtemp(prefix="tc_", dir=grz_tmp("serve"))
                unpacked.append(dst_path)
                test.dump(dst_path)
            relaunch = min(self._relaunch, repeat)
            runner = Runner(
                self.server,
                self.target,
                idle_threshold=idle_threshold,
                idle_delay=idle_delay,
                relaunch=relaunch * test_count,
            )
            # perform iterations
            for _ in range(repeat):
                self.status.iteration += 1
                if self.target.closed:
                    if self._harness is None:
                        location = runner.location(
                            "/grz_current_test", self.server.port
                        )
                    else:
                        location = runner.location(
                            "/grz_harness",
                            self.server.port,
                            close_after=relaunch * test_count,
                            time_limit=time_limit,
                        )
                    startup_error = False
                    # The environment from the initial testcase is used because
                    # a sequence of testcases is expected to be run without
                    # relaunching the Target to match the functionality of
                    # Grizzly. If this is not the case each TestCase should
                    # be run individually.
                    runner.launch(location, env_mod=testcases[0].env_vars)
                # run tests
                durations = list()
                served = list()
                for test_idx in range(test_count):
                    if test_count > 1:
                        LOG.info(
                            "Running test, part %d/%d (%d/%d)...",
                            test_idx + 1,
                            test_count,
                            self.status.iteration,
                            repeat,
                        )
                    else:
                        LOG.info(
                            "Running test (%d/%d)...", self.status.iteration, repeat
                        )
                    # update redirects
                    if self._harness is not None:
                        next_idx = (test_idx + 1) % test_count
                        server_map.set_redirect(
                            "grz_next_test",
                            testcases[next_idx].landing_page,
                            required=True,
                        )
                    server_map.set_redirect(
                        "grz_current_test",
                        testcases[test_idx].landing_page,
                        required=False,
                    )
                    # run testcase
                    run_result = runner.run(
                        self.ignore,
                        server_map,
                        testcases[test_idx],
                        test_path=unpacked[test_idx],
                        wait_for_callback=self._harness is None,
                    )
                    durations.append(run_result.duration)
                    served.append(run_result.served)
                    if run_result.status is not None or not run_result.attempted:
                        break
                if not run_result.attempted:
                    LOG.warning("Test case was not served")
                    if run_result.initial:
                        startup_error = True
                        if run_result.status == RunResult.FAILED:
                            # TODO: what is to best action to take in this case?
                            LOG.warning("Delayed startup failure detected")
                        else:
                            LOG.warning("Timeout too short? System too busy?")
                # process run results
                if run_result.status == RunResult.FAILED:
                    # TODO: use self.target.create_report here
                    log_path = mkdtemp(prefix="logs_", dir=grz_tmp("logs"))
                    self.target.save_logs(log_path)
                    report = Report(
                        log_path, self.target.binary, is_hang=run_result.timeout
                    )
                    # check signatures
                    if run_result.timeout:
                        short_sig = "Potential hang detected"
                    else:
                        short_sig = report.crash_info.createShortSignature()
                    # set active signature
                    if (
                        not startup_error
                        and not self._any_crash
                        and not run_result.timeout
                        and self._signature is None
                        and short_sig != "No crash detected"
                    ):
                        assert not expect_hang
                        LOG.debug("no signature given, using %r", short_sig)
                        self._signature = report.crash_signature
                    # bucket result
                    if short_sig == "No crash detected":
                        # TODO: verify report.major == "NO_STACK"
                        # otherwise FM failed to parse the logs
                        # TODO: change this to support hangs/timeouts, etc
                        LOG.info("Result: No crash detected")
                    elif not startup_error and (
                        self._any_crash
                        or self.check_match(self._signature, report, expect_hang)
                    ):
                        self.status.count_result(short_sig)
                        LOG.info(
                            "Result: %s (%s:%s)",
                            short_sig,
                            report.major[:8],
                            report.minor[:8],
                        )
                        if sig_hash:
                            LOG.debug("using provided signature (hash) to bucket")
                            bucket_hash = sig_hash
                        else:
                            bucket_hash = report.crash_hash
                        if bucket_hash not in reports:
                            reports[bucket_hash] = ReplayResult(
                                report, served, durations, True
                            )
                            LOG.debug("now tracking %s", bucket_hash)
                            report = None  # don't remove report
                        else:
                            reports[bucket_hash].count += 1
                            LOG.debug("already tracking %s", bucket_hash)
                    else:
                        LOG.info(
                            "Result: Different signature: %s (%s:%s)",
                            short_sig,
                            report.major[:8],
                            report.minor[:8],
                        )
                        self.status.ignored += 1
                        if report.crash_hash not in reports:
                            reports[report.crash_hash] = ReplayResult(
                                report, served, durations, False
                            )
                            LOG.debug("now tracking %s", report.crash_hash)
                            report = None  # don't remove report
                        else:
                            reports[report.crash_hash].count += 1
                            LOG.debug("already tracking %s", report.crash_hash)
                    # purge untracked report
                    if report is not None:
                        report.cleanup()
                        report = None
                elif run_result.status == RunResult.IGNORED:
                    self.status.ignored += 1
                    LOG.info("Result: Ignored (%d)", self.status.ignored)

                if exit_early:
                    # failed to meet minimum number of results
                    if (
                        repeat - self.status.iteration + self.status.results
                        < min_results
                    ):
                        if self.status.iteration < repeat:
                            LOG.debug("skipping remaining attempts")
                        # failed to reproduce issue
                        LOG.debug(
                            "results (%d) < minimum (%d), after %d attempts",
                            self.status.results,
                            min_results,
                            self.status.iteration,
                        )
                        # NOTE: this can be tricky if the harness is used because it can
                        # skip the shutdown performed in the harness and runner, if this
                        # is an issue for now use relaunch=1
                        break
                    # check if complete (minimum number of results found)
                    if self.status.results >= min_results:
                        assert self.status.results == min_results
                        assert (
                            sum(x.count for x in reports.values() if x.expected)
                            >= min_results
                        )
                        LOG.debug(
                            "results == expected (%d), after %d attempts",
                            min_results,
                            self.status.iteration,
                        )
                        break

                # TODO: should we warn about large browser logs?

            # process results
            results = list()
            if self._any_crash:
                assert all(x.expected for x in reports.values())
                if sum(x.count for x in reports.values()) >= min_results:
                    results = list(reports.values())
                else:
                    LOG.debug(
                        "%d (any_crash) less than minimum %d",
                        self.status.results,
                        min_results,
                    )
                    for report in reports.values():
                        report.report.cleanup()
            else:
                assert sum(x.expected for x in reports.values()) <= 1
                # filter out unreliable expected results
                for crash_hash, report in reports.items():
                    if report.expected and report.count < min_results:
                        LOG.debug(
                            "%r less than minimum (%d/%d)",
                            crash_hash,
                            report.count,
                            min_results,
                        )
                        report.report.cleanup()
                        continue
                    results.append(report)
            # this should only be displayed when both conditions are met:
            # 1) runner does not close target (no delay was given before shutdown)
            # 2) result has not been successfully reproduced
            if (
                self._relaunch > 1
                and not self.target.closed
                and not any(x.expected for x in results)
            ):
                LOG.info("Perhaps try with --relaunch=1")
            # active reports have been moved to results
            # clear reports to avoid cleanup of active reports
            reports.clear()
            return results

        finally:
            # remove unpacked testcase data
            for tc_path in unpacked:
                rmtree(tc_path)
            self.target.close()
            # remove unprocessed reports
            for report in reports.values():
                report.report.cleanup()

    @classmethod
    def time_limits(cls, time_limit, timeout, tests):
        """Determine the test time limit and timeout. A ConfigError is raised
        if configuration errors are detected.

        Args:
            time_limit (int): Test time limit.
            timeout (int): Iteration timeout.
            tests (iterable): Testcases that may contain time limit values.

        Returns:
            tuple (int, int): Time limit and timeout.
        """
        if time_limit is None:
            test_limits = tuple(int(x.time_limit) for x in tests if x.time_limit)
            if test_limits:
                time_limit = max(test_limits)
            else:
                time_limit = cls.DEFAULT_TIME_LIMIT
        if timeout is None:
            timeout = time_limit + TIMEOUT_DELAY
        if timeout < time_limit:
            raise ConfigError(
                "Timeout (%d) cannot be less than time limit (%d)"
                % (timeout, time_limit),
                Session.EXIT_ARGS,
            )
        return time_limit, timeout

    @classmethod
    def main(cls, args):
        configure_logging(args.log_level)
        if args.fuzzmanager:
            FuzzManagerReporter.sanity_check(args.binary)
            # TODO: add fuzzmanager support

        LOG.info("Starting Grizzly Replay")

        if args.ignore:
            LOG.info("Ignoring: %s", ", ".join(args.ignore))
        if args.xvfb:
            LOG.info("Running with Xvfb")
        if args.rr:
            LOG.info("Running with RR")
        elif args.valgrind:
            LOG.info("Running with Valgrind. This will be SLOW!")

        if args.sig:
            signature = CrashSignature.fromFile(args.sig)
        else:
            signature = None

        try:
            testcases = cls.load_testcases(
                args.input, args.prefs is None, subset=args.test_index
            )
        except TestCaseLoadFailure as exc:
            LOG.error("Error: %s", str(exc))
            return Session.EXIT_ERROR

        results = None
        target = None
        tmp_prefs = None
        try:
            if args.no_harness and len(testcases) > 1:
                LOG.error(
                    "'--no-harness' cannot be used with multiple testcases. "
                    "Perhaps '--test-index' can help."
                )
                return Session.EXIT_ARGS
            # check if hangs are expected
            expect_hang = cls.expect_hang(args.ignore, signature, testcases)
            # check test time limit and timeout
            time_limit, timeout = cls.time_limits(
                args.time_limit, args.timeout, testcases
            )
            LOG.info("Using time limit: %ds, timeout: %ds", time_limit, timeout)
            repeat = max(args.min_crashes, args.repeat)
            if args.no_harness:
                LOG.debug("no-harness enabled, forcing relaunch=1")
                relaunch = 1
            else:
                relaunch = min(args.relaunch, repeat)
            LOG.info(
                "Repeat: %d, Minimum crashes: %d, Relaunch %d",
                repeat,
                args.min_crashes,
                relaunch,
            )
            LOG.debug("initializing the Target")
            target = load_plugin(args.platform, "grizzly_targets", Target)(
                args.binary,
                args.extension,
                args.launch_timeout,
                args.log_limit,
                args.memory,
                rr=args.rr,
                valgrind=args.valgrind,
                xvfb=args.xvfb,
            )
            # prioritize specified prefs.js file over included file
            if args.prefs is not None:
                for testcase in testcases:
                    testcase.add_meta(TestFile.from_file(args.prefs, "prefs.js"))
                LOG.info("Using specified prefs.js")
                target.prefs = args.prefs
            else:
                for testcase in testcases:
                    prefs_tf = testcase.get_file("prefs.js")
                    if prefs_tf:
                        tmp_prefs = mkdtemp(prefix="prefs_", dir=grz_tmp("replay"))
                        prefs_tf.dump(tmp_prefs)
                        LOG.info("Using prefs.js from testcase")
                        target.prefs = pathjoin(tmp_prefs, "prefs.js")
                        break
            LOG.debug("starting sapphire server")
            # launch HTTP server used to serve test cases
            with Sapphire(auto_close=1, timeout=timeout) as server:
                target.reverse(server.port, server.port)
                with cls(
                    args.ignore,
                    server,
                    target,
                    any_crash=args.any_crash,
                    relaunch=relaunch,
                    signature=signature,
                    use_harness=not args.no_harness,
                ) as replay:
                    results = replay.run(
                        testcases,
                        time_limit,
                        expect_hang=expect_hang,
                        idle_delay=args.idle_delay,
                        idle_threshold=args.idle_threshold,
                        min_results=args.min_crashes,
                        repeat=repeat,
                    )
            # handle results
            success = any(x.expected for x in results)
            if success:
                LOG.info("Result successfully reproduced")
            else:
                LOG.info("Failed to reproduce results")
            if args.logs and results:
                cls.report_to_filesystem(
                    args.logs, results, testcases if args.include_test else None
                )
            # TODO: add fuzzmanager reporting
            return Session.EXIT_SUCCESS if success else Session.EXIT_FAILURE

        except ConfigError as exc:
            LOG.error(str(exc))
            return exc.exit_code

        except KeyboardInterrupt:
            return Session.EXIT_ABORT

        except (TargetLaunchError, TargetLaunchTimeout) as exc:
            LOG.error(str(exc))
            if isinstance(exc, TargetLaunchError) and exc.report:
                path = grz_tmp("launch_failures")
                LOG.error("Logs can be found here %r", path)
                reporter = FilesystemReporter(path, major_bucket=False)
                reporter.submit([], exc.report)
            return Session.EXIT_LAUNCH_FAILURE

        finally:
            LOG.info("Shutting down...")
            if results:
                # cleanup unreported results
                for result in results:
                    result.report.cleanup()
            if target is not None:
                target.cleanup()
            for testcase in testcases:
                testcase.cleanup()
            if tmp_prefs is not None:
                rmtree(tmp_prefs, ignore_errors=True)
            LOG.info("Done.")



#!/usr/bin/env python
# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Manage Grizzly status reports."""
from argparse import ArgumentParser
from collections import defaultdict
from datetime import timedelta
from functools import partial
from logging import DEBUG, INFO, basicConfig

try:
    from os import getloadavg
except ImportError:  # pragma: no cover
    # os.getloadavg() is not available on all platforms
    getloadavg = None
from os import SEEK_CUR, getenv, scandir
from os.path import isdir
from re import match
from time import gmtime, strftime, time

from psutil import cpu_count, cpu_percent, disk_usage, virtual_memory

from .status import Status

__all__ = ("StatusReporter",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class StatusReporter:
    """Read and merge Grizzly status reports, including tracebacks if found.
    Output is a single textual report, e.g. for submission to EC2SpotManager.
    """

    CPU_POLL_INTERVAL = 1
    DISPLAY_LIMIT_LOG = 10  # don't include log results unless size exceeds 10MBs
    EXP_LIMIT = 600  # expiration limit, ignore older reports
    READ_BUF_SIZE = 0x10000  # 64KB
    SUMMARY_LIMIT = 4095  # summary output must be no more than 4KB

    def __init__(self, reports, tracebacks=None):
        self.reports = reports
        self.tracebacks = tracebacks

    def dump_specific(self, filename):
        """Write out merged reports.

        Args:
            filename (str): Path where output should be written.

        Returns:
            None
        """
        with open(filename, "w") as ofp:
            ofp.write(self._specific())

    def dump_summary(self, filename, runtime=False, sysinfo=True, timestamp=True):
        """Write out summary merged reports.

        Args:
            filename (str): Path where output should be written.
            runtime (bool): Include total runtime in output
            sysinfo (bool): Include system info (CPU, disk, RAM... etc) in output
            timestamp (bool): Include time stamp in output

        Returns:
            None
        """
        with open(filename, "w") as ofp:
            ofp.write(
                self._summary(runtime=runtime, sysinfo=sysinfo, timestamp=timestamp)
            )

    @property
    def has_results(self):
        return any(x.results for x in self.reports)

    @classmethod
    def load(cls, tb_path=None):
        """Read Grizzly status reports and create a StatusReporter object.

        Args:
            tb_path (str): Directory to scan for files containing Python tracebacks.

        Returns:
            StatusReporter: Contains available status reports and traceback reports.
        """
        tracebacks = None if tb_path is None else cls._tracebacks(tb_path)
        return cls(list(Status.loadall()), tracebacks=tracebacks)

    def print_results(self):
        print(self._results())

    def print_specific(self):
        print(self._specific())

    def print_summary(self, runtime=True, sysinfo=False, timestamp=False):
        print(self._summary(runtime=runtime, sysinfo=sysinfo, timestamp=timestamp))

    def _results(self, max_len=85):
        signatures = defaultdict(int)
        # calculate totals
        for result in self.reports:
            for sig, count in result.signatures():
                signatures[sig] += count
        # generate output
        txt = list()
        for sig, count in sorted(signatures.items(), key=lambda x: x[1], reverse=True):
            if len(sig) > max_len:
                txt.append("%d: '%s...'\n" % (count, sig[:max_len]))
            else:
                txt.append("%d: %r\n" % (count, sig))
        return "".join(txt)

    @staticmethod
    def _scan(path, fname_pattern):
        for entry in scandir(path):
            if match(fname_pattern, entry.name) is None:
                continue
            if not entry.is_file():
                continue
            if entry.stat().st_size:
                yield entry.path

    def _specific(self):
        """Merged and generate formatted output of status reports.

        Args:
            None

        Returns:
            str: A formatted report
        """
        if not self.reports:
            return "No status reports available"
        exp = int(time()) - self.EXP_LIMIT
        self.reports.sort(key=lambda x: x.duration, reverse=True)
        self.reports.sort(key=lambda x: x.timestamp < exp)
        txt = list()
        for num, report in enumerate(self.reports, start=1):
            txt.append("#%02d - %d -" % (num, report.pid))
            if report.timestamp < exp:
                txt.append(" (EXPIRED)\n")
                continue
            txt.append(" Runtime %s\n" % (timedelta(seconds=int(report.duration)),))
            txt.append(" * Iterations: %d" % (report.iteration,))
            txt.append(" @ %0.2f," % (round(report.rate, 2),))
            txt.append(" Ignored: %d," % (report.ignored,))
            txt.append(" Results: %d" % (report.results,))
            txt.append("\n")
            # add profiling data if it exists
            if any(report.profile_entries()):
                txt.append(" * Profiling entries *\n")
            for entry in sorted(
                report.profile_entries(), key=lambda x: x.total, reverse=True
            ):
                avg = entry.total / float(entry.count)
                txt.append(" > %s: %dx " % (entry.name, entry.count))
                if entry.total > 300:
                    txt.append(str(timedelta(seconds=int(entry.total))))
                else:
                    txt.append("%0.3fs" % (round(entry.total, 3),))
                txt.append(
                    " %0.2f%%" % (round(entry.total / report.duration * 100, 2),)
                )
                txt.append(" (%0.3f avg," % (round(avg, 3),))
                txt.append(" %0.3f max," % (round(entry.max, 3),))
                txt.append(" %0.3f min)" % (round(entry.min, 3),))
                txt.append("\n")
        return "".join(txt)

    def _summary(self, runtime=True, sysinfo=False, timestamp=False):
        """Merge and generate a summary of status reports.

        Args:
            filename (str): Path where output should be written.
            runtime (bool): Include total runtime in output
            sysinfo (bool): Include system info (CPU, disk, RAM... etc) in output
            timestamp (bool): Include time stamp in output

        Returns:
            str: A summary of merged reports
        """
        txt = list()
        # Job specific status
        exp = int(time()) - self.EXP_LIMIT
        reports = tuple(x for x in self.reports if x.timestamp > exp)
        if reports:
            # calculate totals
            iterations = tuple(x.iteration for x in reports)
            log_sizes = tuple(x.log_size for x in reports)
            rates = tuple(x.rate for x in reports)
            results = tuple(x.results for x in reports)
            count = len(reports)
            total_ignored = sum(x.ignored for x in reports)
            total_iters = sum(iterations)
            # Iterations
            txt.append("Iterations : %d" % (total_iters,))
            if count > 1:
                txt.append(" (%d, %d)" % (max(iterations), min(iterations)))
            txt.append("\n")
            # Rate
            txt.append("      Rate : %d @ %0.2f" % (count, round(sum(rates), 2)))
            if count > 1:
                txt.append(
                    " (%0.2f, %0.2f)" % (round(max(rates), 2), round(min(rates), 2))
                )
            txt.append("\n")
            # Results / Signature mismatch
            txt.append("   Results : %d" % (sum(results),))
            if total_ignored:
                ignore_pct = total_ignored / float(total_iters) * 100
                txt.append(
                    " (%d ignored @ %0.2f%%)" % (total_ignored, round(ignore_pct, 2))
                )
            # Runtime
            if runtime:
                txt.append("\n")
                total_runtime = sum(x.duration for x in reports)
                txt.append("   Runtime : %s" % (timedelta(seconds=int(total_runtime)),))
            # Log size
            log_usage = sum(log_sizes) / 1048576.0
            if log_usage > self.DISPLAY_LIMIT_LOG:
                txt.append("\n")
                txt.append("      Logs : %0.1fMB" % (log_usage,))
                if count > 1:
                    txt.append(
                        " (%0.2fMB, %0.2fMB)"
                        % (max(log_sizes) / 1048576.0, min(log_sizes) / 1048576.0)
                    )
        else:
            txt.append("No status reports available")
        if sysinfo:
            txt.append("\n")
            txt.append(self._sys_info())
        if timestamp:
            txt.append("\n")
            txt.append(strftime(" Timestamp : %Y/%m/%d %X %z", gmtime()))
        msg = "".join(txt)
        if self.tracebacks:
            txt = self._merge_tracebacks(self.tracebacks, self.SUMMARY_LIMIT - len(msg))
            msg = "".join((msg, txt))
        return msg

    @staticmethod
    def _merge_tracebacks(tracebacks, size_limit):
        """Merge traceback without exceeding size_limit.

        Args:
            tracebacks (iterable): TracebackReport to merge.
            size_limit (int): Maximum size in bytes of output.

        Returns:
            str: merged tracebacks.
        """
        txt = list()
        txt.append("\n\nWARNING Tracebacks (%d) detected!" % (len(tracebacks),))
        tb_size = len(txt[-1])
        for tbr in tracebacks:
            tb_size += len(tbr) + 1
            if tb_size > size_limit:
                break
            txt.append(str(tbr))
        return "\n".join(txt)

    @staticmethod
    def _sys_info():
        """Collect and format system information.

        Args:
            None

        Returns:
            str: System information formatted to match output from _summary()
        """
        txt = list()
        txt.append(
            "CPU & Load : %d @ %0.1f%%"
            % (
                cpu_count(),
                cpu_percent(interval=StatusReporter.CPU_POLL_INTERVAL),
            )
        )
        if getloadavg is not None:
            txt.append(" %s\n" % (str(getloadavg()),))
        else:
            txt.append("\n")
        mem_usage = virtual_memory()
        txt.append("    Memory : ")
        if mem_usage.available < 1073741824:  # < 1GB
            txt.append("%dMB" % (mem_usage.available / 1048576,))
        else:
            txt.append("%0.1fGB" % (mem_usage.available / 1073741824.0,))
        txt.append(" of %0.1fGB free\n" % (mem_usage.total / 1073741824.0,))
        usage = disk_usage("/")
        txt.append("      Disk : ")
        if usage.free < 1073741824:  # < 1GB
            txt.append("%dMB" % (usage.free / 1048576,))
        else:
            txt.append("%0.1fGB" % (usage.free / 1073741824.0,))
        txt.append(" of %0.1fGB free" % (usage.total / 1073741824.0,))
        return "".join(txt)

    @staticmethod
    def _tracebacks(path, ignore_kbi=True, max_preceeding=5):
        """Search screen logs for tracebacks.

        Args:
            path (str): Directory containing log files.
            ignore_kbi (bool): Do not include KeyboardInterupts in results
            max_preceeding (int): Maximum number of lines preceding traceback to
                                  include.

        Returns:
            list: A list of TracebackReports.
        """
        tracebacks = list()
        for screen_log in StatusReporter._scan(path, r"screenlog\.\d+"):
            tbr = TracebackReport.from_file(screen_log, max_preceeding=max_preceeding)
            if tbr is None:
                continue
            if ignore_kbi and tbr.is_kbi:
                continue
            tracebacks.append(tbr)
        return tracebacks


class TracebackReport:
    """Read Python tracebacks from log files and store it in a manner that is helpful
    when generating reports.
    """

    MAX_LINES = 16  # should be no less than 6
    READ_LIMIT = 0x10000  # 64KB

    def __init__(self, file_name, lines, is_kbi=False, prev_lines=None):
        assert isinstance(lines, list)
        self.file_name = file_name
        self.lines = lines
        self.prev_lines = list() if prev_lines is None else prev_lines
        self.is_kbi = is_kbi

    @classmethod
    def from_file(cls, input_log, max_preceeding=5):
        """Create TracebackReport from a text file containing a Python traceback.
        Only the first traceback in the file will be parsed.

        Args:
            input_log (str): File to parse.
            max_preceeding (int): Number of lines to collect leading up to the
                                  traceback.

        Returns:
            TracebackReport: Contains data from input_log.
        """
        token = b"Traceback (most recent call last):"
        assert len(token) < cls.READ_LIMIT
        try:
            with open(input_log, "rb") as in_fp:
                for chunk in iter(partial(in_fp.read, cls.READ_LIMIT), b""):
                    idx = chunk.find(token)
                    if idx > -1:
                        # calculate offset of data in the file
                        pos = in_fp.tell() - len(chunk) + idx
                        break
                    if len(chunk) == cls.READ_LIMIT:
                        # seek back to avoid missing beginning of token
                        in_fp.seek(len(token) * -1, SEEK_CUR)
                else:
                    # no traceback here, move along
                    return None
                # seek back 2KB to collect preceding lines
                in_fp.seek(max(pos - 2048, 0))
                data = in_fp.read(cls.READ_LIMIT)
        except OSError:  # pragma: no cover
            # in case the file goes away
            return None

        data = data.decode("ascii", errors="ignore").splitlines()
        token = token.decode()
        is_kbi = False
        tb_start = None
        tb_end = None
        line_count = len(data)
        for line_num, log_line in enumerate(data):
            if tb_start is None and token in log_line:
                tb_start = line_num
                continue
            if tb_start is not None:
                log_line = log_line.strip()
                if not log_line:
                    # stop at first empty line
                    tb_end = min(line_num, line_count)
                    break
                if match(r"^\w+(\.\w+)*\:\s|^\w+(Interrupt|Error)$", log_line):
                    is_kbi = log_line.startswith("KeyboardInterrupt")
                    # stop after error message
                    tb_end = min(line_num + 1, line_count)
                    break
        assert tb_start is not None
        if max_preceeding > 0:
            prev_start = max(tb_start - max_preceeding, 0)
            prev_lines = data[prev_start:tb_start]
        else:
            prev_lines = None
        if tb_end is None:
            # limit if the end is not identified (failsafe)
            tb_end = max(line_count, cls.MAX_LINES)
        if tb_end - tb_start > cls.MAX_LINES:
            # add first entry
            lines = data[tb_start : tb_start + 3]
            lines += ["<--- TRACEBACK TRIMMED--->"]
            # add end entries
            lines += data[tb_end - (cls.MAX_LINES - 3) : tb_end]
        else:
            lines = data[tb_start:tb_end]
        return cls(input_log, lines, is_kbi=is_kbi, prev_lines=prev_lines)

    def __len__(self):
        return len(str(self))

    def __str__(self):
        return "\n".join(["Log: %r" % self.file_name] + self.prev_lines + self.lines)


def main(args=None):
    """Merge Grizzly status files into a single report (main entrypoint).

    Args:
        args (list/None): Argument list to parse instead of sys.argv (for testing).

    Returns:
        None
    """
    if bool(getenv("DEBUG")):  # pragma: no cover
        log_level = DEBUG
        log_fmt = "%(levelname).1s %(name)s [%(asctime)s] %(message)s"
    else:
        log_level = INFO
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt="%Y-%m-%d %H:%M:%S", level=log_level)

    modes = ("status",)
    parser = ArgumentParser(description="Grizzly status report generator")
    parser.add_argument("--dump", help="File to write report to")
    parser.add_argument(
        "--mode",
        default="status",
        help="Status loading mode. Available modes: %s (default: 'status')"
        % (", ".join(modes),),
    )
    parser.add_argument(
        "--system-report",
        action="store_true",
        help="Output summary and system information",
    )
    parser.add_argument(
        "--tracebacks",
        help="Scan path for Python tracebacks found in screenlog.# files",
    )
    args = parser.parse_args(args)
    if args.mode not in modes:
        parser.error("Invalid mode %r" % args.mode)
    if args.tracebacks and not isdir(args.tracebacks):
        parser.error("--tracebacks must be a directory")

    reporter = StatusReporter.load(tb_path=args.tracebacks)
    if args.dump:
        reporter.dump_summary(args.dump)
        return 0
    if not reporter.reports:
        print("Grizzly Status - No status reports to display")
        return 0
    print(
        "Grizzly Status - %s - Instance report frequency: %ds\n"
        % (strftime("%Y/%m/%d %X"), Status.REPORT_FREQ)
    )
    print("[Reports]")
    reporter.print_specific()
    if reporter.has_results:
        print("[Result Signatures]")
        reporter.print_results()
    print("[Summary]")
    reporter.print_summary(sysinfo=args.system_report)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import DEBUG, getLogger
from os import getcwd, getpid
from os.path import join as pathjoin

from sapphire import Sapphire

from .adapter import Adapter
from .common.plugins import load as load_plugin
from .common.reporter import (
    FilesystemReporter,
    FuzzManagerReporter,
    S3FuzzManagerReporter,
)
from .common.utils import TIMEOUT_DELAY, configure_logging
from .session import Session
from .target import Target, TargetLaunchError, TargetLaunchTimeout

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


LOG = getLogger(__name__)


def main(args):
    configure_logging(args.log_level)
    LOG.info("Starting Grizzly (%d)", getpid())
    if args.fuzzmanager:
        FuzzManagerReporter.sanity_check(args.binary)
    elif args.s3_fuzzmanager:
        S3FuzzManagerReporter.sanity_check(args.binary)

    if args.ignore:
        LOG.info("Ignoring: %s", ", ".join(args.ignore))
    if args.xvfb:
        LOG.info("Running with Xvfb")
    if args.rr:
        LOG.info("Running with RR")
    if args.valgrind:
        LOG.info("Running with Valgrind. This will be SLOW!")

    adapter = None
    session = None
    target = None
    try:
        LOG.debug("initializing Adapter %r", args.adapter)
        adapter = load_plugin(args.adapter, "grizzly_adapters", Adapter)(args.adapter)

        # test time limit and timeout sanity checking
        if args.time_limit:
            time_limit = args.time_limit
        else:
            assert adapter.TIME_LIMIT >= 1
            time_limit = adapter.TIME_LIMIT
        if args.timeout:
            timeout = args.timeout
        else:
            timeout = time_limit + TIMEOUT_DELAY
        LOG.info("Using test time limit: %ds, timeout: %ds", time_limit, timeout)
        if timeout < time_limit:
            LOG.error("Timeout must be at least test time limit if not greater")
            return Session.EXIT_ARGS
        if adapter.HARNESS_FILE and time_limit == timeout:
            LOG.warning(
                "To avoid relaunches due to tests failing to close"
                " themselves use a timeout greater than time limit"
            )

        if adapter.RELAUNCH > 0:
            LOG.info("Relaunch (%d) set in Adapter", adapter.RELAUNCH)
            relaunch = adapter.RELAUNCH
        else:
            relaunch = args.relaunch

        LOG.debug("initializing the Target %r", args.platform)
        target = load_plugin(args.platform, "grizzly_targets", Target)(
            args.binary,
            args.extension,
            args.launch_timeout,
            args.log_limit,
            args.memory,
            rr=args.rr,
            valgrind=args.valgrind,
            xvfb=args.xvfb,
        )
        if args.prefs:
            target.prefs = args.prefs
            LOG.info("Using prefs %r", args.prefs)
        adapter.monitor = target.monitor

        LOG.debug("initializing the Reporter")
        if args.fuzzmanager:
            LOG.info("Results will be reported via FuzzManager")
            reporter = FuzzManagerReporter(tool=args.tool)
        elif args.s3_fuzzmanager:
            LOG.info(
                "Results will be reported via FuzzManager w/ large attachments in S3"
            )
            reporter = S3FuzzManagerReporter(tool=args.tool)
        else:
            reporter = FilesystemReporter(pathjoin(getcwd(), "results"))
            LOG.info("Results will be stored in %r", reporter.report_path)

        if args.limit > 0:
            LOG.info("%r iteration(s) will be attempted", args.limit)

        # set 'auto_close=1' so the client error pages (code 4XX) will
        # call 'window.close()' after a second.
        # launch http server used to serve test cases
        LOG.debug("starting Sapphire server")
        with Sapphire(auto_close=1, timeout=timeout) as server:
            target.reverse(server.port, server.port)
            LOG.debug("initializing the Session")
            session = Session(
                adapter,
                reporter,
                server,
                target,
                coverage=args.coverage,
                enable_profiling=args.enable_profiling,
                relaunch=relaunch,
                report_size=args.collect,
            )
            if args.log_level == DEBUG or args.verbose:
                display_mode = Session.DISPLAY_VERBOSE
            else:
                display_mode = Session.DISPLAY_NORMAL
            session.run(
                args.ignore,
                time_limit,
                input_path=args.input,
                iteration_limit=args.limit,
                display_mode=display_mode,
            )

    except KeyboardInterrupt:
        LOG.info("Ctrl+C detected.")
        return Session.EXIT_ABORT

    except (TargetLaunchError, TargetLaunchTimeout) as exc:
        LOG.error(str(exc))
        return Session.EXIT_LAUNCH_FAILURE

    finally:
        LOG.info("Shutting down...")
        if session is not None:
            LOG.debug("calling session.close()")
            session.close()
        if target is not None:
            LOG.debug("calling target.cleanup()")
            target.cleanup()
        if adapter is not None:
            LOG.debug("calling adapter.cleanup()")
            adapter.cleanup()
        LOG.info("Done.")

    return Session.EXIT_SUCCESS



# coding=utf-8
"""
Sapphire HTTP server
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .core import Sapphire
from .job import SERVED_ALL, SERVED_NONE, SERVED_REQUEST, SERVED_TIMEOUT
from .server_map import ServerMap

__all__ = (
    "Sapphire",
    "SERVED_ALL",
    "SERVED_NONE",
    "SERVED_REQUEST",
    "SERVED_TIMEOUT",
    "ServerMap",
)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from grizzly.common import TestCase

from . import NoOpAdapter


def test_no_op_01():
    """test a simple Adapter"""
    adapter = NoOpAdapter("no-op")
    adapter.setup(None, None)
    test = TestCase("a", "b", adapter.name)
    assert not test.data_size
    assert not test.contains("a")
    adapter.generate(test, None)
    assert test.contains("a")



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from argparse import ArgumentParser, HelpFormatter
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from os.path import exists, isdir, isfile
from platform import system

from .common.plugins import scan as scan_plugins
from .common.utils import TIMEOUT_DELAY


# ref: https://stackoverflow.com/questions/12268602/sort-argparse-help-alphabetically
class SortingHelpFormatter(HelpFormatter):
    @staticmethod
    def __sort_key(action):
        for opt in action.option_strings:
            if opt.startswith("--"):
                return [opt]
        return action.option_strings

    def add_usage(self, usage, actions, groups, prefix=None):
        actions = sorted(actions, key=self.__sort_key)
        super().add_usage(usage, actions, groups, prefix)

    def add_arguments(self, actions):
        actions = sorted(actions, key=self.__sort_key)
        super().add_arguments(actions)


class CommonArgs:
    IGNORABLE = ("log-limit", "memory", "timeout")
    IGNORE = ("log-limit", "timeout")

    def __init__(self):
        # log levels for console logging
        self._level_map = {
            "CRIT": CRITICAL,
            "ERROR": ERROR,
            "WARN": WARNING,
            "INFO": INFO,
            "DEBUG": DEBUG,
        }
        self._sanity_skip = set()

        if not hasattr(self, "parser"):
            self.parser = ArgumentParser(
                formatter_class=SortingHelpFormatter, conflict_handler="resolve"
            )

        self.parser.add_argument("binary", help="Firefox binary to run")
        self.parser.add_argument(
            "--log-level",
            default="INFO",
            help="Configure console logging. Options: %s (default: %%(default)s)"
            % ", ".join(
                k for k, v in sorted(self._level_map.items(), key=lambda x: x[1])
            ),
        )

        self.launcher_grp = self.parser.add_argument_group("Launcher Arguments")
        self.launcher_grp.add_argument(
            "-e",
            "--extension",
            action="append",
            help="Install an extension. Specify the path to the xpi or the directory"
            " containing the unpacked extension. To install multiple extensions"
            " specify multiple times",
        )
        self.launcher_grp.add_argument(
            "--launch-timeout",
            type=int,
            default=300,
            help="Number of seconds to wait before LaunchError is raised"
            " (default: %(default)s)",
        )
        self.launcher_grp.add_argument(
            "--log-limit",
            type=int,
            default=0,
            help="Browser log file size limit in MBs (default: 'no limit')",
        )
        self.launcher_grp.add_argument(
            "-m",
            "--memory",
            type=int,
            default=0,
            help="Browser process memory limit in MBs (default: 'no limit')",
        )
        self.launcher_grp.add_argument(
            "--platform",
            default="ffpuppet",
            help="Installed Platforms (Targets): %s (default: %%(default)s)"
            % ", ".join(sorted(scan_plugins("grizzly_targets"))),
        )
        self.launcher_grp.add_argument("-p", "--prefs", help="prefs.js file to use")
        self.launcher_grp.add_argument(
            "--relaunch",
            type=int,
            default=1000,
            help="Number of iterations performed before relaunching the browser"
            " (default: %(default)s)",
        )
        self.launcher_grp.add_argument(
            "--rr", action="store_true", help="Use rr (Linux only)"
        )
        self.launcher_grp.add_argument(
            "--time-limit",
            type=int,
            default=None,
            help="This is the maximum amount of time that a test is expected to take."
            " After the time has elapsed the harness will attempt to close the test."
            " By default `Adapter.TIME_LIMIT` is used."
            " Browser build types and debuggers can affect the amount of time"
            " required to run a test case.",
        )
        self.launcher_grp.add_argument(
            "-t",
            "--timeout",
            type=int,
            default=None,
            help="Iteration timeout in seconds. By default this is `test-duration`+%ds."
            " If the timeout is reached the target is assumed to be in a bad state"
            " and will be closed. Typically this should be a few seconds greater"
            " than the value used for `test-duration`." % (TIMEOUT_DELAY,),
        )
        self.launcher_grp.add_argument(
            "--valgrind", action="store_true", help="Use Valgrind (Linux only)"
        )
        self.launcher_grp.add_argument(
            "--xvfb", action="store_true", help="Use Xvfb (Linux only)"
        )

        self.reporter_grp = self.parser.add_argument_group("Reporter Arguments")
        self.reporter_grp.add_argument(
            "--fuzzmanager", action="store_true", help="Report results to FuzzManager"
        )
        self.reporter_grp.add_argument(
            "--ignore",
            nargs="*",
            default=list(self.IGNORE),
            help="Space separated list of issue types to ignore. Valid options: %s"
            " (default: %s)" % (" ".join(self.IGNORABLE), " ".join(self.IGNORE)),
        )
        self.reporter_grp.add_argument(
            "--tool",
            help="Override tool name used when reporting issues to FuzzManager",
        )

        self.parser.epilog = (
            "For addition help check out the wiki:"
            " https://github.com/MozillaSecurity/grizzly/wiki"
        )

    def parse_args(self, argv=None):
        args = self.parser.parse_args(argv)
        self.sanity_check(args)
        return args

    def sanity_check(self, args):
        targets = scan_plugins("grizzly_targets")
        if not targets:
            self.parser.error("No Platforms (Targets) are installed")

        if "binary" not in self._sanity_skip and not isfile(args.binary):
            self.parser.error("file not found: %r" % (args.binary,))

        # sanitize ignore list
        args.ignore = {arg.lower() for arg in args.ignore}
        for ignore in args.ignore:
            if ignore not in self.IGNORABLE:
                self.parser.error("Unrecognized ignore value %r" % (ignore,))

        # check log level
        log_level = self._level_map.get(args.log_level.upper(), None)
        if log_level is None:
            self.parser.error("Invalid log-level %r" % (args.log_level,))
        args.log_level = log_level

        if args.log_limit < 0:
            self.parser.error("--log-limit must be >= 0")
        args.log_limit *= 1_048_576

        if args.memory < 0:
            self.parser.error("--memory must be >= 0")
        args.memory *= 1_048_576

        if args.relaunch < 1:
            self.parser.error("--relaunch must be >= 1")

        if args.extension:
            for ext in args.extension:
                if not exists(ext):
                    self.parser.error("%r does not exist" % (ext,))
                if not isdir(ext) or (isfile(ext) and ext.endswith(".xpi")):
                    self.parser.error("Extension must be a folder or .xpi")

        if args.platform not in targets:
            self.parser.error("Platform %r not installed" % (args.platform,))

        if args.prefs and not isfile(args.prefs):
            self.parser.error("--prefs file not found")

        if args.rr and args.valgrind:
            self.parser.error("--rr and --valgrind are mutually exclusive")

        if args.time_limit is not None and args.time_limit < 1:
            self.parser.error("--time-limit must be >= 1")

        if args.timeout is not None and args.timeout < 1:
            self.parser.error("--timeout must be >= 1")

        if "tool" not in self._sanity_skip:
            if args.tool is not None and not args.fuzzmanager:
                self.parser.error("--tool can only be given with --fuzzmanager")

        # check platform specific args
        os_name = system()
        if args.rr and os_name != "Linux":
            self.parser.error("--rr is only supported on Linux")
        if args.valgrind and os_name != "Linux":
            self.parser.error("--valgrind is only supported on Linux")
        if args.xvfb and os_name != "Linux":
            self.parser.error("--xvfb is only supported on Linux")


class GrizzlyArgs(CommonArgs):
    def __init__(self):
        super().__init__()
        self._sanity_skip.add("tool")
        self.parser.add_argument(
            "adapter",
            help="Installed Adapters: %s"
            % ", ".join(sorted(scan_plugins("grizzly_adapters"))),
        )
        self.parser.add_argument(
            "--enable-profiling",
            action="store_true",
            help="Record profiling data. The data can be viewed by running the"
            " status reporter while running Grizzly.",
        )
        self.parser.add_argument(
            "-i", "--input", help="Test case or directory containing test cases"
        )
        self.parser.add_argument(
            "--limit",
            type=int,
            default=0,
            help="Maximum number of iterations to be performed. (default: no limit)",
        )
        self.parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="Output console updates every iteration. By default the number"
            " of iterations between console updates doubles each update."
            " Updates are always printed when a result is detected or the"
            " target is relaunched.",
        )

        self.launcher_grp.add_argument(
            "--coverage", action="store_true", help="Enable coverage collection"
        )

        self.reporter_grp.add_argument(
            "-c",
            "--collect",
            type=int,
            default=1,
            help="Maximum number of test cases to include in the report"
            "(default: %(default)s)",
        )
        self.reporter_grp.add_argument(
            "--s3-fuzzmanager",
            action="store_true",
            help="Report large attachments (if any) to S3 and then the crash &"
            " S3 link to FuzzManager",
        )

    def sanity_check(self, args):
        super().sanity_check(args)
        adapters = scan_plugins("grizzly_adapters")
        if not adapters:
            self.parser.error("No Adapters are installed")

        if args.adapter not in adapters:
            self.parser.error("Adapter %r is not installed" % (args.adapter,))

        if args.collect < 1:
            self.parser.error("--collect must be greater than 0")

        if args.fuzzmanager and args.s3_fuzzmanager:
            self.parser.error(
                "--fuzzmanager and --s3-fuzzmanager are mutually exclusive"
            )

        if args.input and not exists(args.input):
            self.parser.error("%r does not exist" % (args.input,))

        if args.limit < 0:
            self.parser.error("--limit must be >= 0 (0 = no limit)")

        if args.tool is not None and not (args.fuzzmanager or args.s3_fuzzmanager):
            self.parser.error(
                "--tool can only be given with --fuzzmanager/--s3-fuzzmanager"
            )



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
unit tests for grizzly.replay.args
"""
from pytest import mark, raises

from .args import ReplayArgs


def test_replay_args_01(capsys, mocker, tmp_path):
    """test parsing args"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=["targ1"])
    # missing args tests
    with raises(SystemExit):
        ReplayArgs().parse_args([])
    # specified prefs.js missing
    exe = tmp_path / "binary"
    exe.touch()
    # test success
    ReplayArgs().parse_args([str(exe), str(exe), "--platform", "targ1"])
    # test missing input
    with raises(SystemExit):
        ReplayArgs().parse_args(argv=[str(exe), "missing", "--platform", "targ1"])
    assert "error: 'missing' does not exist" in capsys.readouterr()[-1]


@mark.parametrize(
    "args, msg",
    [
        # test any-crash with signature
        (
            ["--any-crash", "--sig", "x"],
            "error: signature is ignored when running with --any-crash",
        ),
        # test in valid idle_delay
        (
            ["--idle-threshold", "1", "--idle-delay", "-1"],
            "error: --idle-delay value must be positive",
        ),
        # test invalid min-crashes value
        (["--min-crashes", "0"], "error: --min-crashes value must be positive"),
        # test invalid repeat value
        (["--repeat", "-1"], "error: --repeat value must be positive"),
        # test missing signature file
        (["--sig", "missing"], "error: signature file not found"),
    ],
)
def test_replay_args_02(capsys, mocker, tmp_path, args, msg):
    """test CommonArgs.parse_args() - sanity checks"""
    target = "target1"
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=[target])
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    with raises(SystemExit):
        ReplayArgs().parse_args(
            argv=[str(fake_bin), str(fake_bin), "--platform", target] + args
        )
    assert msg in capsys.readouterr()[-1]



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""`ReduceManager` finds the smallest testcase(s) to reproduce an issue."""
import json
import os
from itertools import chain
from locale import LC_ALL, setlocale
from logging import getLogger
from math import ceil, log
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from time import time

from FTB.Signatures.CrashInfo import CrashSignature

from sapphire import Sapphire

from ..common.fuzzmanager import CrashEntry
from ..common.plugins import load as load_plugin
from ..common.reporter import FilesystemReporter, FuzzManagerReporter
from ..common.storage import TestCaseLoadFailure, TestFile
from ..common.utils import ConfigError, configure_logging, grz_tmp
from ..replay import ReplayManager
from ..session import Session
from ..target import Target, TargetLaunchError, TargetLaunchTimeout
from .exceptions import GrizzlyReduceBaseException, NotReproducible
from .stats import ReductionStats
from .strategies import STRATEGIES

__author__ = "Jesse Schwartzentruber"
__credits__ = ["Jesse Schwartzentruber", "Tyson Smith"]


LOG = getLogger(__name__)


class ReduceManager:
    """Manage reduction of one or more testcases to find the smallest testcase
    that reproduces a given issue.

    Attributes:
        ignore (list(str)): Classes of results to ignore (see `--ignore`).
        server (sapphire.Sapphire): Server instance to serve testcases.
        strategies (list(str)): List of strategies to use for reducing
                                testcases (in order).
        target (grizzly.target.Target): Target instance to run testcases.
        testcases (list(grizzly.common.storage.TestCase)): List of one or more Grizzly
                                                           testcases to reduce.
    """

    ANALYSIS_ITERATIONS = 11  # number of iterations to analyze
    # --min-crashes value when analysis is used and reliability is less than perfect
    ANALYSIS_MIN_CRASHES = 1
    ANALYSIS_PERFECT_MIN_CRASHES = 2  # --min-crashes when reliability is perfect
    # probability that successful reduction will observe the crash
    ANALYSIS_TARGET_PROBABILITY = 0.95
    # to see the worst case, run the `repeat` calculation in run_reliability_analysis
    # using `crashes_percent = 1.0/ANALYSIS_ITERATIONS`

    IDLE_DELAY_MIN = 10
    IDLE_DELAY_DURATION_MULTIPLIER = 1.5
    ITER_TIMEOUT_MIN = 10
    ITER_TIMEOUT_DURATION_MULTIPLIER = 2

    def __init__(
        self,
        ignore,
        server,
        target,
        testcases,
        strategies,
        log_path,
        any_crash=False,
        expect_hang=False,
        idle_delay=0,
        idle_threshold=0,
        relaunch=1,
        report_period=None,
        report_to_fuzzmanager=False,
        signature=None,
        signature_desc=None,
        static_timeout=False,
        tool=None,
        use_analysis=True,
        use_harness=True,
    ):
        """Initialize reduction manager. Many arguments are common with `ReplayManager`.

        Args:
            ignore (list(str)): Value for `self.ignore` attribute.
            server (sapphire.Sapphire): Value for `self.server` attribute.
            target (grizzly.target.Target): Value for `self.target` attribute.
            testcases (list(grizzly.common.storage.TestCase)):
                Value for `self.testcases` attribute.
            strategies (list(str)): Value for `self.strategies` attribute.
            log_path (Path or str): Path to save results when reporting to filesystem.
            any_crash (bool): Accept any crash when reducing, not just those matching
                              the specified or first observed signature.
            expect_hang (bool): Attempt to reduce a test that triggers a hang.
            idle_delay (int): Number of seconds to wait before polling for idle.
            idle_threshold (int): CPU usage threshold to mark the process as idle.
            relaunch (int): Maximum number of iterations performed by Runner
                            before Target should be relaunched.
            report_period (int or None): Periodically report best results for
                                         long-running strategies.
            report_to_fuzzmanager (bool): Report to FuzzManager rather than filesystem.
            signature (FTB.Signatures.CrashInfo.CrashSignature or None):
                Signature for accepting crashes.
            signature_desc (str): Short description of the given signature.
            static_timeout (bool): Use only specified timeouts (`--timeout` and
                                   `--idle-delay`), even if testcase appears to need
                                   less time.
            tool (str or None): Override tool when reporting to FuzzManager.
            use_analysis (bool): Analyse reliability of testcase before running each
                                 reduction strategy.
            use_harness (bool): Whether to allow use of harness when navigating
                                between testcases.
        """
        self.ignore = ignore
        self.server = server
        self.strategies = strategies
        self.target = target
        self.testcases = testcases
        self._any_crash = any_crash
        self._expect_hang = expect_hang
        self._idle_delay = idle_delay
        self._idle_threshold = idle_threshold
        self._log_path = Path(log_path) if isinstance(log_path, str) else log_path
        # these parameters may be overwritten during analysis, so keep a copy of them
        self._original_relaunch = relaunch
        self._original_use_harness = use_harness
        self._report_to_fuzzmanager = report_to_fuzzmanager
        self._report_periodically = report_period
        self._report_tool = tool
        self._signature = signature
        self._signature_desc = signature_desc
        self._static_timeout = expect_hang or static_timeout
        self._stats = ReductionStats()
        self._use_analysis = use_analysis
        self._use_harness = use_harness

    def update_timeout(self, results):
        """Tune idle/server timeout values based on actual duration of expected results.

        Expected durations will be updated if the actual duration is much lower.

        Timeouts are not updated in three cases:

            - `static_timeout=True` is passed to constructor (`--static-timeout`),
            - `any_crash=True` is passed to constructor (`--any-crash`),
            - Target is running under valgrind (`--valgrind`).

        Arguments:
            results (grizzly.replay.ReplayResult):
                Observed results. Any given expected results may affect the idle delay
                and sapphire timeout.

        Returns:
            None
        """
        # TODO: properly handle test duration and timeout
        assert self._static_timeout or not self._expect_hang
        if (
            self._static_timeout
            or self._any_crash
            or getattr(self.target, "use_valgrind", False)
        ):
            # the amount of time it can take to replay a test case can vary
            # when under Valgrind so do not update the timeout in that case

            # when any_crash is given, crashes may be completely unrelated (all are
            # expected), so lowering timeout or idle delay will only hide crashes
            return

        durations = list(
            chain.from_iterable(
                result.durations for result in results if result.expected
            )
        )
        if not durations:
            # no expected results
            return
        run_time = max(durations)

        # If `run_time * multiplier` is less than idle poll delay, update it
        LOG.debug("Run time %r", run_time)
        new_idle_delay = max(
            self.IDLE_DELAY_MIN,
            min(run_time * self.IDLE_DELAY_DURATION_MULTIPLIER, self._idle_delay),
        )
        if new_idle_delay < self._idle_delay:
            LOG.info("Updating poll delay to: %r", new_idle_delay)
            self._idle_delay = new_idle_delay
        # If `run_time * multiplier` is less than iter_timeout, update it
        # in other words, decrease the timeout if this ran in less than half the timeout
        new_iter_timeout = max(
            self.ITER_TIMEOUT_MIN,
            min(run_time * self.ITER_TIMEOUT_DURATION_MULTIPLIER, self.server.timeout),
        )
        if new_iter_timeout < self.server.timeout:
            LOG.info("Updating max timeout to: %r", new_iter_timeout)
            self.server.timeout = new_iter_timeout

    def run_reliability_analysis(self, stats):
        """Run several analysis passes of the current testcase to find `run` parameters.

        The number of repetitions and minimum number of crashes are calculated to
        maximize the chances of observing the expected crash.

        Arguments:
            stats (object): Open stats timer (from `ReductionStats.add_timed`).
                            Pass to `ReductionStats.copy` to properly terminate
                            the timer for reporting.

        Returns:
            tuple(int, int, dict): Values for `repeat` and `min_crashes` resulting from
                                   analysis, and info to include in stats.
        """
        harness_last_crashes = 0
        harness_crashes = 0
        non_harness_crashes = 0
        info = {}

        # Reset parameters.
        # Use repeat=1 & relaunch=ITERATIONS because this is closer to how we will run
        #   post-analysis.
        # We're only using repeat=1 instead of repeat=ITERATIONS so we can get feedback
        #   on every call to interesting.

        # N.B. We only use `last_test_only` if `len(self.testcases) > 1` ..
        # if `self.testcases` only has 1 entry to begin with, we don't need
        # `last_test_only` to trim it
        for (use_harness, last_test_only) in (
            (True, True),
            (True, False),
            # only one of the two use_harness=False cases will run:
            # input is len(self.testcases)>1 and we will only try the last testcase
            (False, True),
            # input is len(self.testcases)==1 already and there's no need to trim
            (False, False),
        ):
            if use_harness and (not self._original_use_harness or harness_crashes):
                # Don't test with harness again if we already found crashes with the
                # harness (last_test_only) or if it was disabled by command-line.
                continue
            if not use_harness and harness_crashes >= self.ANALYSIS_ITERATIONS / 2:
                # Don't test without harness if harness found > 50% crashes
                continue
            if last_test_only and len(self.testcases) == 1:
                # Only set `last_test_only` if we initially have more than one testcase
                continue
            if not use_harness and (not last_test_only and len(self.testcases) > 1):
                # Can't run without harness if we have more than one testcase
                # (`last_test_only` will run)
                continue

            if use_harness and (last_test_only or len(self.testcases) == 1):
                relaunch = self.ANALYSIS_ITERATIONS
            else:
                relaunch = 1

            with ReplayManager(
                self.ignore,
                self.server,
                self.target,
                any_crash=self._any_crash,
                relaunch=relaunch,
                signature=self._signature,
                use_harness=use_harness,
            ) as replay:
                LOG.info(
                    "Running for %d iterations to assess reliability %s harness.",
                    self.ANALYSIS_ITERATIONS,
                    "using" if use_harness else "without",
                )
                testcases = self.testcases
                if last_test_only:
                    if use_harness:
                        LOG.warning("Checking reliability with only the last testcase.")
                    else:
                        LOG.warning(
                            "Only the last testcase of %d given will be used to "
                            "assess reliability without harness.",
                            len(testcases),
                        )
                    testcases = [testcases[-1]]
                results = replay.run(
                    testcases,
                    self.server.timeout,
                    repeat=self.ANALYSIS_ITERATIONS,
                    min_results=1,
                    exit_early=False,
                    idle_delay=self._idle_delay,
                    idle_threshold=self._idle_threshold,
                )
                try:
                    stats.add_iterations(replay.status.iteration)
                    crashes = sum(x.count for x in results if x.expected)
                    if crashes and not self._any_crash and self._signature_desc is None:
                        first_expected = next(
                            (report for report in results if report.expected), None
                        )
                        self._signature_desc = (
                            first_expected.report.crash_info.createShortSignature()
                        )
                    self.report(
                        [result for result in results if not result.expected],
                        testcases,
                        self._stats.copy(stats),
                    )
                    if use_harness:
                        # set harness_crashes in both cases (last_test True/False)
                        # we only want to iterate through all testcases if the last
                        # testcase alone never reproduced (crashes == 0).
                        harness_crashes = crashes
                        if last_test_only:
                            harness_last_crashes = crashes
                    else:
                        non_harness_crashes = crashes
                finally:
                    for result in results:
                        result.report.cleanup()
                reliability = crashes / self.ANALYSIS_ITERATIONS
                key = ("using" if use_harness else "without") + " harness"
                if last_test_only:
                    key += "/last test only"
                LOG.info(
                    "Testcase was interesting %0.1f%% of %d attempts %s.",
                    100.0 * reliability,
                    self.ANALYSIS_ITERATIONS,
                    key,
                )
                info[key] = reliability
                # ensure same signature is always used
                self._signature = replay.signature

        if not (harness_crashes or non_harness_crashes):
            raise NotReproducible("Did not reproduce during analysis")

        # if harness is selected, we'll only use the last testcase
        if harness_last_crashes:
            harness_crashes = harness_last_crashes

        # should we use the harness? go with harness unless no-harness crashed 50% more
        self._use_harness = not (
            non_harness_crashes > harness_crashes
            and (
                harness_crashes == 0
                or (non_harness_crashes - harness_crashes) / harness_crashes >= 0.5
            )
        )

        if (self._use_harness and harness_last_crashes) or (
            not self._use_harness and len(self.testcases) > 1
        ):
            LOG.warning(
                "Last testcase %s harness was selected, other %d "
                "testcases in the original will be ignored.",
                "with" if self._use_harness else "without",
                len(self.testcases) - 1,
            )
            while len(self.testcases) > 1:
                self.testcases.pop(0).cleanup()

        crashes_percent = (
            harness_crashes if self._use_harness else non_harness_crashes
        ) / self.ANALYSIS_ITERATIONS

        # adjust repeat/min-crashes depending on how reliable the testcase was
        if abs(crashes_percent - 1) < 0.01:
            min_crashes = self.ANALYSIS_PERFECT_MIN_CRASHES
        else:
            min_crashes = self.ANALYSIS_MIN_CRASHES
        # crashes_percent is max 99.9% to avoid domain errors
        repeat = int(
            ceil(
                log(
                    1 - self.ANALYSIS_TARGET_PROBABILITY,
                    1 - min(crashes_percent, 0.9999),
                )
            )
            * min_crashes
        )

        LOG.info("Analysis results:")
        if harness_crashes >= self.ANALYSIS_ITERATIONS / 2:
            LOG.info(
                "* testcase was better than 50% reliable with the harness "
                "(--no-harness not assessed)"
            )
        elif harness_crashes == non_harness_crashes:
            LOG.info("* testcase was equally reliable with/without the harness")
        elif not self._original_use_harness:
            LOG.info("* --no-harness was already set")
        else:
            LOG.info(
                "* testcase was %s reliable with the harness",
                "more" if harness_crashes > non_harness_crashes else "less",
            )
        return (repeat, min_crashes, info)

    def testcase_size(self):
        """Calculate the current testcase size.

        Returns:
            int: Current size of the testcase(s).
        """
        return sum(tc.data_size for tc in self.testcases)

    def run(self, repeat=1, min_results=1):
        """Run testcase reduction.

        Args:
            repeat (int): Maximum number of times to run the TestCase.
            min_results (int): Minimum number of results needed before run can
                               be considered successful.

        Returns:
            int: One of the `Session.EXIT_*` constants.
        """
        any_success = False
        sig_given = self._signature is not None
        last_reports = None
        last_tried = None
        self._stats.add("init", self.testcase_size())
        # record total stats overall so that any time missed by individual milestones
        # will still be included in the total
        with self._stats.add_timed("final", self.testcase_size) as total_stats:
            if self._use_analysis:
                with total_stats.add_timed("analysis") as stats:
                    (
                        repeat,
                        min_results,
                        reliability_info,
                    ) = self.run_reliability_analysis(stats)
                self._stats.add_info("Analysis", reliability_info)
                any_success = True  # analysis ran and didn't raise
            # multi part test cases should always use relaunch == 1
            # since that can mean a delay is required
            if self._use_harness and len(self.testcases) == 1:
                relaunch = min(self._original_relaunch, repeat)
            else:
                relaunch = 1
            LOG.info(
                "Repeat: %d, Minimum crashes: %d, Relaunch %d",
                repeat,
                min_results,
                relaunch,
            )
            self._stats.add_info(
                "Run Parameters",
                {
                    "Repeat": repeat,
                    "Min Crashes": min_results,
                    "Relaunch": relaunch,
                    "Harness": self._use_harness,
                },
            )

            for strategy_no, strategy in enumerate(self.strategies):
                LOG.info("")
                LOG.info(
                    "Using strategy %s (%d/%d)",
                    strategy,
                    strategy_no + 1,
                    len(self.strategies),
                )
                replay = ReplayManager(
                    self.ignore,
                    self.server,
                    self.target,
                    any_crash=self._any_crash,
                    relaunch=relaunch,
                    signature=self._signature,
                    use_harness=self._use_harness,
                )
                strategy = STRATEGIES[strategy](self.testcases)
                if last_tried is not None:
                    strategy.update_tried(last_tried)
                    last_tried = None

                strategy_last_report = time()
                strategy_stats = total_stats.add_timed(strategy.name)
                best_results = []
                try:
                    with replay, strategy, strategy_stats:
                        for reduction in strategy:
                            keep_reduction = False
                            results = []
                            try:
                                # reduction is a new list of testcases to be
                                # replayed
                                results = replay.run(
                                    reduction,
                                    self.server.timeout,
                                    expect_hang=self._expect_hang,
                                    idle_delay=self._idle_delay,
                                    idle_threshold=self._idle_threshold,
                                    min_results=min_results,
                                    repeat=repeat,
                                )
                                strategy_stats.add_iterations(replay.status.iteration)
                                strategy_stats.add_attempts(1)
                                self.update_timeout(results)
                                # get the first expected result (if any),
                                #   and update the strategy
                                first_expected = next(
                                    (report for report in results if report.expected),
                                    None,
                                )
                                success = first_expected is not None
                                if success:
                                    strategy_stats.add_successes(1)
                                    if (
                                        not self._any_crash
                                        and self._signature_desc is None
                                    ):
                                        self._signature_desc = (
                                            # pylint: disable=line-too-long
                                            first_expected.report.crash_info.createShortSignature()  # noqa: E501
                                        )
                                served = None
                                if success and not self._any_crash:
                                    served = first_expected.served
                                strategy.update(success, served=served)
                                if strategy.name == "check" and not success:
                                    raise NotReproducible("Not reproducible at 'check'")
                                any_success = any_success or success
                                # if the reduction reproduced,
                                #   update self.testcases (new best)
                                if success:
                                    LOG.info("Reduction succeeded")
                                    for testcase in self.testcases:
                                        testcase.cleanup()
                                    self.testcases = reduction
                                    keep_reduction = True
                                    # cleanup old best results
                                    for result in best_results:
                                        result.report.cleanup()
                                    # filter expected results out into `best_results`
                                    best_results = [
                                        result for result in results if result.expected
                                    ]
                                    results = [
                                        result
                                        for result in results
                                        if not result.expected
                                    ]
                                else:
                                    LOG.info("Attempt failed")
                                # if the reduction found other crashes,
                                #   report those immediately
                                self.report(
                                    results, reduction, self._stats.copy(strategy_stats)
                                )

                                now = time()
                                if (
                                    self._report_periodically
                                    and best_results
                                    and now - strategy_last_report
                                    > self._report_periodically
                                ):
                                    last_reports = self.report(
                                        best_results,
                                        self.testcases,
                                        self._stats.copy(strategy_stats),
                                    )
                                    for result in best_results:
                                        result.report.cleanup()
                                    best_results = []
                                    strategy_last_report = now
                                    LOG.info("Best results reported (periodic)")

                            finally:
                                if not keep_reduction:
                                    for testcase in reduction:
                                        testcase.cleanup()
                                for result in results:
                                    result.report.cleanup()

                        # if self._signature was already set, this will do nothing
                        # otherwise, ensure the first found signature is used throughout
                        self._signature = replay.signature

                    if best_results:
                        last_reports = self.report(
                            best_results, self.testcases, self._stats.copy(total_stats)
                        )

                except KeyboardInterrupt:
                    if best_results:
                        last_reports = self.report(
                            best_results, self.testcases, self._stats.copy(total_stats)
                        )
                        LOG.warning(
                            "Ctrl+C detected, best reduction so far " "reported as %r",
                            last_reports,
                        )
                    raise
                finally:
                    for result in best_results:
                        result.report.cleanup()

                # store "tried" cache to pass to next strategy
                last_tried = strategy.get_tried()

            # if we complete all strategies, mark the last reported crashes as reduced
            if self._report_to_fuzzmanager and last_reports:
                for crash_id in last_reports:
                    LOG.info(
                        "Updating crash %d to quality %s",
                        crash_id,
                        FuzzManagerReporter.quality_name(
                            FuzzManagerReporter.QUAL_REDUCED_RESULT
                        ),
                    )
                    CrashEntry(
                        crash_id
                    ).testcase_quality = FuzzManagerReporter.QUAL_REDUCED_RESULT

        # it's possible we made it this far without ever setting signature_desc.
        # this is only possible if --no-analysis is given
        # just give None instead of trying to format the CrashSignature
        self._stats.add_info(
            "Signature",
            {
                "given": sig_given,
                "any": self._any_crash,
                "description": self._signature_desc,
            },
        )

        # log a summary of what was done.
        LOG.info(
            "Reduction summary:%s%s",
            os.linesep,
            os.linesep.join(self._stats.format_lines()),
        )

        if any_success:
            return Session.EXIT_SUCCESS
        return Session.EXIT_FAILURE

    def report(self, results, testcases, stats=None):
        """Report results, either to FuzzManager or to filesystem.

        Arguments:
            results (list(ReplayResult)): Results observed during reduction.
            testcases (list(TestCase)): Testcases used to trigger results.
            stats (ReductionStats): Statistics for reduction of these results.
                                    (may be different than `self._stats`)

        Returns:
            list(*): List of return values from `reporter.submit()`.
        """
        ret_values = []
        for result in results:
            if self._report_to_fuzzmanager:
                reporter = FuzzManagerReporter(self._report_tool)
                if result.expected:
                    reporter.force_report = True
            else:
                report_dir = "reports" if result.expected else "other_reports"
                reporter = FilesystemReporter(
                    report_path=str(self._log_path / report_dir), major_bucket=False
                )
            # write reduction stats
            if stats is not None:
                report_path = Path(result.report.path)
                with (report_path / "reduce_stats.txt").open("w") as out:
                    for line in stats.format_lines():
                        print(line, file=out)
                (report_path / "reduce_stats.json").write_text(stats.json())
                if self._report_to_fuzzmanager:
                    stats.add_to_reporter(reporter)
            # clone the tests so we can safely call purge_optional here for each report
            # (report.served may be different for non-expected or any-crash results)
            clones = [test.clone() for test in testcases]
            try:
                if result.served is not None:
                    for clone, served in zip(clones, result.served):
                        clone.purge_optional(served)
                ret_values.append(reporter.submit(clones, report=result.report))
            finally:
                for clone in clones:
                    clone.cleanup()
        return ret_values

    @classmethod
    def main(cls, args):
        """CLI for `grizzly.reduce`.

        Arguments:
            args (argparse.Namespace): Result from `ReduceArgs.parse_args`.

        Returns:
            int: 0 for success. non-0 indicates a problem.
        """
        # pylint: disable=too-many-return-statements
        configure_logging(args.log_level)
        setlocale(LC_ALL, "")
        if args.fuzzmanager:
            FuzzManagerReporter.sanity_check(args.binary)

        LOG.info("Starting Grizzly Reduce")

        if args.ignore:
            LOG.info("Ignoring: %s", ", ".join(args.ignore))
        if args.xvfb:
            LOG.info("Running with Xvfb")
        if args.valgrind:
            LOG.info("Running with Valgrind. This will be SLOW!")
        if args.rr:
            LOG.info("Running with RR")

        signature = None
        signature_desc = None
        target = None
        testcases = []
        tmp_prefs = None
        try:
            if args.sig:
                signature = CrashSignature.fromFile(args.sig)
                meta = Path(args.sig).with_suffix(".metadata")
                if meta.is_file():
                    meta = json.loads(meta.read_text())
                    signature_desc = meta["shortDescription"]

            try:
                testcases = ReplayManager.load_testcases(
                    args.input, args.prefs is None, subset=args.test_index
                )
            except TestCaseLoadFailure as exc:
                LOG.error("Error: %s", str(exc))
                return Session.EXIT_ERROR

            if args.tool is None and testcases[0].adapter_name is not None:
                LOG.warning(
                    "Setting default --tool=grizzly-%s from testcase",
                    testcases[0].adapter_name,
                )
                args.tool = "grizzly-%s" % (testcases[0].adapter_name,)

            expect_hang = ReplayManager.expect_hang(args.ignore, signature, testcases)

            if args.no_harness:
                if len(testcases) > 1:
                    LOG.error(
                        "Error: '--no-harness' cannot be used with multiple "
                        "testcases. Perhaps '--test-index' can help."
                    )
                    return Session.EXIT_ARGS
                LOG.debug("--no-harness specified relaunch set to 1")
                args.relaunch = 1

            # check test time limit and timeout
            # TODO: add support for test time limit, use timeout in both cases for now
            _, timeout = ReplayManager.time_limits(
                args.timeout, args.timeout, testcases
            )

            args.repeat = max(args.min_crashes, args.repeat)
            relaunch = min(args.relaunch, args.repeat)
            LOG.debug("initializing the Target")
            target = load_plugin(args.platform, "grizzly_targets", Target)(
                args.binary,
                args.extension,
                args.launch_timeout,
                args.log_limit,
                args.memory,
                rr=args.rr,
                valgrind=args.valgrind,
                xvfb=args.xvfb,
            )
            # prioritize specified prefs.js file over included file
            if args.prefs is not None:
                for testcase in testcases:
                    testcase.add_meta(TestFile.from_file(args.prefs, "prefs.js"))
                LOG.info("Using specified prefs.js")
                target.prefs = args.prefs
            else:
                for testcase in testcases:
                    prefs_tf = testcase.get_file("prefs.js")
                    if prefs_tf:
                        tmp_prefs = Path(
                            mkdtemp(prefix="prefs_", dir=grz_tmp("replay"))
                        )
                        prefs_tf.dump(str(tmp_prefs))
                        LOG.info("Using prefs.js from testcase")
                        target.prefs = str(tmp_prefs / "prefs.js")
                        break

            LOG.debug("starting sapphire server")
            # launch HTTP server used to serve test cases
            with Sapphire(auto_close=1, timeout=timeout) as server:
                target.reverse(server.port, server.port)
                mgr = ReduceManager(
                    args.ignore,
                    server,
                    target,
                    testcases,
                    args.strategies,
                    args.logs,
                    any_crash=args.any_crash,
                    expect_hang=expect_hang,
                    idle_delay=args.idle_delay,
                    idle_threshold=args.idle_threshold,
                    relaunch=relaunch,
                    report_period=args.report_period,
                    report_to_fuzzmanager=args.fuzzmanager,
                    signature=signature,
                    signature_desc=signature_desc,
                    static_timeout=args.static_timeout,
                    tool=args.tool,
                    use_analysis=not args.no_analysis,
                    use_harness=not args.no_harness,
                )
                return_code = mgr.run(repeat=args.repeat, min_results=args.min_crashes)
            return return_code

        except ConfigError as exc:
            LOG.error(str(exc))
            return exc.exit_code

        except KeyboardInterrupt as exc:
            LOG.error("Exception: %r", exc)
            return Session.EXIT_ABORT

        except (TargetLaunchError, TargetLaunchTimeout) as exc:
            LOG.error("Exception: %s", exc)
            if isinstance(exc, TargetLaunchError) and exc.report:
                path = grz_tmp("launch_failures")
                LOG.error("Logs can be found here %r", path)
                reporter = FilesystemReporter(path, major_bucket=False)
                reporter.submit([], exc.report)
            return Session.EXIT_LAUNCH_FAILURE

        except GrizzlyReduceBaseException as exc:
            LOG.error(exc.msg)
            return exc.code

        except Exception:  # pylint: disable=broad-except
            LOG.exception("Exception during reduction!")
            return Session.EXIT_ERROR

        finally:
            LOG.info("Shutting down...")
            if target is not None:
                target.cleanup()
            for testcase in testcases:
                testcase.cleanup()
            if tmp_prefs is not None:
                rmtree(str(tmp_prefs), ignore_errors=True)
            LOG.info("Done.")



# coding: utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from random import choice, getrandbits
from struct import pack
from tempfile import SpooledTemporaryFile

from pytest import mark, raises

from .args import parse_args
from .loki import Loki


@mark.parametrize(
    "in_size, aggression, byte_order",
    [
        (1, 0.1, ">"),
        (2, 0.1, "<"),
        (3, 0.1, None),
        (4, 0.1, None),
        (5, 0.5, None),
        (100, 0.2, None),
    ],
)
def test_loki_fuzz_file(tmp_path, in_size, aggression, byte_order):
    """test Loki.fuzz_file() with different file sizes"""
    in_data = b"A" * in_size
    tmp_fn = tmp_path / "input"
    tmp_fn.write_bytes(in_data)

    out_path = tmp_path / "out"
    out_path.mkdir()

    fuzzer = Loki(aggression=aggression, byte_order=byte_order)
    for _ in range(100):
        assert fuzzer.fuzz_file(str(tmp_fn), 1, str(out_path))
        out_files = list(out_path.iterdir())
        assert len(out_files) == 1
        out_data = out_files[0].read_bytes()
        assert len(out_data) == in_size
        if out_data != in_data:
            break
    else:
        raise AssertionError("failed to fuzz data")


def test_loki_01(tmp_path):
    """test Loki.fuzz_file() error cases"""
    fuzzer = Loki(aggression=0.1)
    # test missing file
    assert not fuzzer.fuzz_file("nofile.test", 1, str(tmp_path))
    assert not list(tmp_path.iterdir())
    # test empty file
    tmp_fn = tmp_path / "input"
    tmp_fn.touch()
    out_path = tmp_path / "out"
    out_path.mkdir()
    assert not fuzzer.fuzz_file(str(tmp_fn), 1, str(out_path))
    assert not list(out_path.iterdir())


def test_loki_02():
    """test Loki.fuzz_data()"""
    in_data = b"This is test DATA!"
    in_size = len(in_data)

    fuzzer = Loki(aggression=0.1)
    for _ in range(100):
        out_data = fuzzer.fuzz_data(in_data)
        assert len(out_data) == in_size
        if in_data not in out_data:
            break
    else:
        raise AssertionError("failed to fuzz data")


def test_loki_fuzz_01(mocker):
    """test Loki._fuzz()"""
    loki = Loki(aggression=1)
    # test empty input sample
    with SpooledTemporaryFile(max_size=10, mode="r+b") as tmp_fp:
        loki._fuzz(tmp_fp)
    # test multiple mutations
    with SpooledTemporaryFile(max_size=10, mode="r+b") as tmp_fp:
        tmp_fp.write(b"12345678")
        loki._fuzz(tmp_fp)
    # make remaining tests deterministic
    mocker.patch("loki.loki.getrandbits", autospec=True, return_value=1)
    mocker.patch("loki.loki.randint", autospec=True, side_effect=min)
    mocker.patch("loki.loki.sample", autospec=True, return_value=[0])
    # test multi-byte with > 3 bytes
    with SpooledTemporaryFile(max_size=10, mode="r+b") as tmp_fp:
        tmp_fp.write(b"1234")
        loki._fuzz(tmp_fp)
    # test multi-byte with > 1 byte
    with SpooledTemporaryFile(max_size=10, mode="r+b") as tmp_fp:
        tmp_fp.write(b"12")
        loki._fuzz(tmp_fp)
    # test multi-byte with 1 byte and byte order set
    loki.byte_order = ">"
    with SpooledTemporaryFile(max_size=10, mode="r+b") as tmp_fp:
        tmp_fp.write(b"1")
        loki._fuzz(tmp_fp)


def test_loki_fuzz_02(mocker):
    """test Loki._fuzz_data() paths"""
    fake_randint = mocker.patch("loki.loki.randint", autospec=True)
    # fuzz op 0
    fake_randint.side_effect = (0, 1, 1)
    Loki._fuzz_data(b"1", "<")
    # fuzz op 1
    fake_randint.side_effect = (1, 0)
    Loki._fuzz_data(b"1", "<")
    # fuzz op 2
    fake_randint.side_effect = (2,)
    Loki._fuzz_data(b"1", ">")
    # fuzz op 3
    fake_randint.side_effect = (3,)
    Loki._fuzz_data(b"1", ">")
    # fuzz op 4 & test data size 1
    fake_randint.side_effect = max
    Loki._fuzz_data(b"1", "<")
    # fuzz op 4 & test data size 2
    Loki._fuzz_data(b"12", "<")
    # fuzz op 4 & test data size 4
    fake_randint.side_effect = (4, 1)
    Loki._fuzz_data(b"1234", ">")
    # invalid data size
    with raises(AssertionError, match=r"Unsupported data size:"):
        Loki._fuzz_data(b"", ">")


def test_loki_stress_01():
    """test Loki._fuzz_data() with random input"""
    orders = ("<", ">")
    sizes = (1, 2, 4)
    for _ in range(3000):
        size = choice(sizes)
        if size == 1:
            in_data = pack("B", getrandbits(8))
        elif size == 2:
            in_data = pack("H", getrandbits(16))
        elif size == 4:
            in_data = pack("I", getrandbits(32))
        assert len(Loki._fuzz_data(in_data, choice(orders))) == size


def test_main_01(mocker, tmp_path):
    """test main()"""
    out_path = tmp_path / "out"
    out_path.mkdir()
    # no output path provided
    fake_mkdtemp = mocker.patch(
        "loki.loki.mkdtemp", autospec=True, return_value=str(out_path)
    )
    sample = tmp_path / "file.bin"
    sample.write_bytes(b"test!")
    args = mocker.Mock(
        aggression=0.1, byte_order=None, count=15, input=str(sample), output=None
    )
    assert Loki.main(args) == 0
    assert fake_mkdtemp.call_count == 1


def test_args_01(capsys):
    """test parse_args()"""
    assert parse_args(argv=["sample"])
    # invalid byte order
    with raises(SystemExit):
        parse_args(argv=["sample", "-b", "a"])
    assert "Invalid byte order" in capsys.readouterr()[1]



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from pkg_resources import EntryPoint
from pytest import raises

from .plugins import PluginLoadError, load, scan


class FakeType1:
    pass


class FakeType2:
    pass


def test_load_01(mocker):
    """test load() - nothing to load"""
    mocker.patch(
        "grizzly.common.plugins.iter_entry_points", autospec=True, return_value=[]
    )
    with raises(PluginLoadError, match="'test-name' not found in 'test-group'"):
        load("test-name", "test-group", FakeType1)


def test_load_02(mocker):
    """test load() - successful load"""
    entry = mocker.Mock(set_spec=EntryPoint)
    entry.name = "test-name"
    entry.load.return_value = FakeType1
    mocker.patch(
        "grizzly.common.plugins.iter_entry_points", autospec=True, return_value=[entry]
    )
    assert load("test-name", "test-group", FakeType1)


def test_load_03(mocker):
    """test load() - invalid type"""
    entry = mocker.Mock(set_spec=EntryPoint)
    entry.name = "test-name"
    entry.load.return_value = FakeType1
    mocker.patch(
        "grizzly.common.plugins.iter_entry_points", autospec=True, return_value=[entry]
    )
    with raises(PluginLoadError, match="'test-name' doesn't inherit from FakeType2"):
        load("test-name", "test-group", FakeType2)


def test_scan_01(mocker):
    """test scan() - no entries found"""
    mocker.patch(
        "grizzly.common.plugins.iter_entry_points", autospec=True, return_value=[]
    )
    assert not scan("test_group")


def test_scan_02(mocker):
    """test scan() - duplicate entry"""
    entry = mocker.Mock(set_spec=EntryPoint)
    entry.name = "test_entry"
    mocker.patch(
        "grizzly.common.plugins.iter_entry_points",
        autospec=True,
        return_value=[entry, entry],
    )
    with raises(PluginLoadError, match="Duplicate entry 'test_entry' in 'test_group'"):
        scan("test_group")


def test_scan_03(mocker):
    """test scan() - success"""
    entry = mocker.Mock(set_spec=EntryPoint)
    entry.name = "test-name"
    mocker.patch(
        "grizzly.common.plugins.iter_entry_points",
        autospec=True,
        return_value=[entry],
    )
    assert "test-name" in scan("test_group")



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABCMeta, abstractmethod
from os import walk
from os.path import abspath, dirname, isdir, isfile
from os.path import join as pathjoin

__all__ = ("Adapter", "AdapterError")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class AdapterError(Exception):
    """The base class for exceptions raised by an Adapter"""


class Adapter(metaclass=ABCMeta):
    """An Adapter is an interface between Grizzly and a fuzzer. A subclass must
    be created in order to add support for additional fuzzers. The Adapter is
    responsible for handling input/output data and executing the fuzzer.
    It is expected that any processes launched or file created on file system
    in the adapter will also be cleaned up in the adapter.
    NOTE: Some methods must not be overloaded doing so will prevent Grizzly from
    operating correctly.

    Attributes:
        _harness (str): Path to harness file that will be used. If None, no
                        harness will be used.
        fuzz (dict): Available as a safe scratch pad for the end-user.
        monitor (TargetMonitor): Used to provide Target status information to
                                 the adapter.
        name (str): Name of the adapter.
        remaining (int): Can be used to indicate the number of TestCases
                         remaining to process.
    """

    HARNESS_FILE = pathjoin(dirname(__file__), "..", "common", "harness.html")
    # Only report test cases with served content.
    IGNORE_UNSERVED = True
    # Maximum iterations between Target relaunches (<1 use default)
    RELAUNCH = 0
    # Maximum execution time per test (used as minimum timeout). The iteration is
    # expected to be complete. If the test is still open the harness will attempt to
    # close it.
    TIME_LIMIT = 30

    __slots__ = ("_harness", "fuzz", "monitor", "name", "remaining")

    def __init__(self, name):
        assert isinstance(name, str)
        if not name:
            raise AdapterError("name must not be empty")
        self._harness = None
        self.fuzz = dict()
        self.monitor = None
        self.name = name
        self.remaining = None

    def cleanup(self):
        """Automatically called once at shutdown. Used internally by Grizzly.
        *** DO NOT OVERLOAD! ***

        Args:
            None

        Returns:
            None
        """
        self.shutdown()

    def enable_harness(self, file_path=None):
        """Enable use of a harness during fuzzing. By default no harness is used.
        *** DO NOT OVERLOAD! ***

        Args:
            file_path (str): Path to file to use as a harness. If None the default
                             harness is used.

        Returns:
            None
        """
        if file_path is None:
            file_path = self.HARNESS_FILE
        with open(file_path, "rb") as in_fp:
            self._harness = in_fp.read()

    def get_harness(self):
        """Get the harness. Used internally by Grizzly.
        *** DO NOT OVERLOAD! ***

        Args:
            None

        Returns:
            TestFile: The active harness.
        """
        return self._harness

    @staticmethod
    def scan_path(path, ignore=("desktop.ini", "thumbs.db"), recursive=False):
        """Scan a path and yield the files within it. This is available as
        a helper method.

        Args:
            path (str): Path to file or directory.
            ignore (iterable(str)): Files to ignore.
            recursive (bool): Scan recursively into directories.

        Yields:
            str: Absolute path to files.
        """
        full_path = abspath(path)
        if isdir(full_path):
            for root, _, files in walk(full_path):
                for fname in files:
                    if fname in ignore or fname.startswith("."):
                        # skip ignored and hidden system files
                        continue
                    yield pathjoin(root, fname)
                if not recursive:
                    break
        elif isfile(full_path):
            yield full_path

    @abstractmethod
    def generate(self, testcase, server_map):
        """Automatically called. Populate testcase here.

        Args:
            testcase (TestCase): TestCase intended to be populated.
            server_map (ServerMap): A ServerMap.

        Returns:
            None
        """

    def on_served(self, testcase, served):
        """Optional. Automatically called after a test case is successfully served.

        Args:
            testcase (TestCase): TestCase that was served.
            served (list(str)): Files served from testcase.

        Returns:
            None
        """

    def on_timeout(self, testcase, served):
        """Optional. Automatically called if timeout occurs while attempting to
        serve a test case. By default it calls `self.on_served()`.

        Args:
            testcase (TestCase): TestCase that was served.
            served (list(str)): Files served from testcase.

        Returns:
            None
        """
        self.on_served(testcase, served)

    def pre_launch(self):
        """Optional. Automatically called before launching the Target.

        Args:
            None

        Returns:
            None
        """

    def setup(self, input_path, server_map):
        """Optional. Automatically called once at startup.

        Args:
            input_path (str): Points to a file or directory passed by the user.
                              None is passed by default.
            server_map (ServerMap): A ServerMap

        Returns:
            None
        """

    def shutdown(self):
        """Optional. Automatically called once at shutdown.

        Args:
            None

        Returns:
            None
        """



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly main"""
from pytest import mark

from sapphire import Sapphire

from .adapter import Adapter
from .main import main
from .session import Session
from .target import Target, TargetLaunchError


class FakeArgs:
    def __init__(self):
        self.binary = None
        self.input = None
        self.adapter = None
        self.collect = 1
        self.coverage = False
        self.enable_profiling = False
        self.extension = None
        self.fuzzmanager = False
        self.ignore = list()
        self.launch_timeout = 300
        self.limit = 0
        self.log_level = 10  # 10 = DEBUG, 20 = INFO
        self.log_limit = 0
        self.memory = 0
        self.platform = "fake-target"
        self.prefs = None
        self.rr = False  # pylint: disable=invalid-name
        self.relaunch = 1000
        self.s3_fuzzmanager = False
        self.time_limit = None
        self.timeout = None
        self.tool = None
        self.valgrind = False
        self.verbose = False
        self.xvfb = False


@mark.parametrize(
    "cov, adpt_relaunch, limit, verbose",
    [
        # successful run
        (False, 0, 0, True),
        # successful run (with limit)
        (False, 0, 10, True),
        # successful run (with coverage)
        (True, 0, 0, False),
        # relaunch 1
        (False, 1, 0, False),
        # relaunch 10
        (False, 10, 0, False),
    ],
)
def test_main_01(mocker, cov, adpt_relaunch, limit, verbose):
    """test main()"""
    fake_adapter = mocker.NonCallableMock(spec_set=Adapter)
    fake_adapter.RELAUNCH = adpt_relaunch
    fake_adapter.TIME_LIMIT = 10
    fake_target = mocker.NonCallableMock(spec_set=Target)
    plugin_loader = mocker.patch("grizzly.main.load_plugin", autospec=True)
    plugin_loader.side_effect = (
        mocker.Mock(spec_set=Adapter, return_value=fake_adapter),
        mocker.Mock(spec_set=Target, return_value=fake_target),
    )
    fake_session = mocker.patch("grizzly.main.Session", autospec=True)
    fake_session.return_value.server = mocker.Mock(spec_set=Sapphire)
    fake_session.EXIT_SUCCESS = Session.EXIT_SUCCESS
    args = FakeArgs()
    args.adapter = "fake"
    args.ignore = ["fake", "fake"]
    args.limit = limit
    args.prefs = "fake"
    args.rr = True
    args.valgrind = True
    args.xvfb = True
    args.verbose = verbose
    if not verbose:
        args.log_level = 20
    args.coverage = cov
    assert main(args) == Session.EXIT_SUCCESS
    assert fake_session.mock_calls[0][-1]["coverage"] == cov
    if adpt_relaunch:
        assert fake_session.mock_calls[0][-1]["relaunch"] == adpt_relaunch
    else:
        # check default
        assert fake_session.mock_calls[0][-1]["relaunch"] == 1000
    assert fake_session.return_value.run.call_count == 1
    assert fake_target.cleanup.call_count == 1


@mark.parametrize(
    "reporter",
    [
        # Default reporter
        None,
        # FuzzManager Reporter
        "FuzzManager",
        # S3FuzzManager Reporter
        "S3FuzzManager",
    ],
)
def test_main_02(mocker, reporter):
    """test main() - test reporters"""
    fake_adapter = mocker.NonCallableMock(spec_set=Adapter)
    fake_adapter.RELAUNCH = 0
    fake_adapter.TIME_LIMIT = 10
    fake_target = mocker.NonCallableMock(spec_set=Target)
    plugin_loader = mocker.patch("grizzly.main.load_plugin", autospec=True)
    plugin_loader.side_effect = (
        mocker.Mock(spec_set=Adapter, return_value=fake_adapter),
        mocker.Mock(spec_set=Target, return_value=fake_target),
    )
    fake_session = mocker.patch("grizzly.main.Session", autospec=True)
    fake_session.return_value.server = mocker.Mock(spec_set=Sapphire)
    fake_session.EXIT_SUCCESS = Session.EXIT_SUCCESS
    args = FakeArgs()
    args.adapter = "fake"
    if reporter == "FuzzManager":
        fake_reporter = mocker.patch("grizzly.main.FuzzManagerReporter", autospec=True)
        fake_reporter.sanity_check.return_value = True
        args.fuzzmanager = True
    elif reporter == "S3FuzzManager":
        fake_reporter = mocker.patch(
            "grizzly.main.S3FuzzManagerReporter", autospec=True
        )
        fake_reporter.sanity_check.return_value = True
        args.s3_fuzzmanager = True
    assert main(args) == Session.EXIT_SUCCESS
    assert fake_target.cleanup.call_count == 1


@mark.parametrize(
    "exit_code, to_raise",
    [
        # test user abort
        (Session.EXIT_ABORT, KeyboardInterrupt()),
        # test launch failure
        (Session.EXIT_LAUNCH_FAILURE, TargetLaunchError("test", None)),
    ],
)
def test_main_03(mocker, exit_code, to_raise):
    """test main() - exit codes"""
    fake_adapter = mocker.NonCallableMock(spec_set=Adapter, name="fake")
    fake_adapter.RELAUNCH = 0
    fake_adapter.TIME_LIMIT = 10
    fake_target = mocker.NonCallableMock(spec_set=Target)
    plugin_loader = mocker.patch("grizzly.main.load_plugin", autospec=True)
    plugin_loader.side_effect = (
        mocker.Mock(spec_set=Adapter, return_value=fake_adapter),
        mocker.Mock(spec_set=Target, return_value=fake_target),
    )
    fake_session = mocker.patch("grizzly.main.Session", autospec=True)
    fake_session.EXIT_SUCCESS = Session.EXIT_SUCCESS
    fake_session.EXIT_ABORT = Session.EXIT_ABORT
    fake_session.EXIT_ARGS = fake_session.EXIT_ARGS = Session.EXIT_ARGS
    fake_session.EXIT_LAUNCH_FAILURE = Session.EXIT_LAUNCH_FAILURE
    fake_session.return_value.server = mocker.Mock(spec_set=Sapphire)
    args = FakeArgs()
    args.adapter = "fake"
    args.input = "fake"
    fake_session.return_value.run.side_effect = to_raise
    assert main(args) == exit_code
    assert fake_target.cleanup.call_count == 1


@mark.parametrize(
    "arg_testlimit, arg_timeout, exit_code",
    [
        # use default test time limit and timeout values
        (None, None, Session.EXIT_SUCCESS),
        # set test time limit
        (10, None, Session.EXIT_SUCCESS),
        # set both test time limit and timeout to the same value
        (10, 10, Session.EXIT_SUCCESS),
        # set timeout greater than test time limit
        (10, 11, Session.EXIT_SUCCESS),
        # set test time limit greater than timeout
        (11, 10, Session.EXIT_ARGS),
    ],
)
def test_main_04(mocker, arg_testlimit, arg_timeout, exit_code):
    """test main() - time-limit and timeout"""
    fake_adapter = mocker.NonCallableMock(spec_set=Adapter, name="fake")
    fake_adapter.RELAUNCH = 1
    fake_adapter.TIME_LIMIT = 10
    fake_target = mocker.NonCallableMock(spec_set=Target)
    plugin_loader = mocker.patch("grizzly.main.load_plugin", autospec=True)
    plugin_loader.side_effect = (
        mocker.Mock(spec_set=Adapter, return_value=fake_adapter),
        mocker.Mock(spec_set=Target, return_value=fake_target),
    )
    fake_session = mocker.patch("grizzly.main.Session", autospec=True)
    fake_session.return_value.server = mocker.Mock(spec_set=Sapphire)
    fake_session.EXIT_ARGS = Session.EXIT_ARGS
    fake_session.EXIT_SUCCESS = Session.EXIT_SUCCESS
    args = FakeArgs()
    args.adapter = "fake"
    args.time_limit = arg_testlimit
    args.timeout = arg_timeout
    assert main(args) == exit_code
